from re import Match
from typing import Any, Dict

from typing_extensions import override

from aws_lambda_powertools.event_handler import ApiGatewayResolver
from aws_lambda_powertools.event_handler.api_gateway import (
    ProxyEventType,
    ResponseBuilder,
)
from aws_lambda_powertools.utilities.data_classes import BedrockAgentEvent


class BedrockResponseBuilder(ResponseBuilder):
    """
    Bedrock Response Builder. This builds the response dict to be returned by Lambda when using Bedrock Agents.

    Since the payload format is different from the standard API Gateway Proxy event, we override the build method.
    """

    @override
    def build(self, event: BedrockAgentEvent, *args) -> Dict[str, Any]:
        """Build the full response dict to be returned by the lambda"""
        self._route(event, None)

        body = self.response.body
        if self.response.is_json() and not isinstance(self.response.body, str):
            body = self.serializer(self.response.body)

        return {
            "messageVersion": "1.0",
            "response": {
                "actionGroup": event.action_group,
                "apiPath": event.api_path,
                "httpMethod": event.http_method,
                "httpStatusCode": self.response.status_code,
                "responseBody": {
                    self.response.content_type: {
                        "body": body,
                    },
                },
            },
        }


class BedrockAgentResolver(ApiGatewayResolver):
    """Bedrock Agent Resolver

    See https://aws.amazon.com/bedrock/agents/ for more information.

    Examples
    --------
    Simple example with a custom lambda handler using the Tracer capture_lambda_handler decorator

    ```python
    from aws_lambda_powertools import Tracer
    from aws_lambda_powertools.event_handler import BedrockAgentResolver

    tracer = Tracer()
    app = BedrockAgentResolver()

    @app.get("/claims")
    def simple_get():
        return "You have 3 claims"

    @tracer.capture_lambda_handler
    def lambda_handler(event, context):
        return app.resolve(event, context)
    """

    current_event: BedrockAgentEvent

    def __init__(self, debug: bool = False, enable_validation: bool = True):
        super().__init__(
            proxy_type=ProxyEventType.BedrockAgentEvent,
            cors=None,
            debug=debug,
            serializer=None,
            strip_prefixes=None,
            enable_validation=enable_validation,
        )
        self._response_builder_class = BedrockResponseBuilder

    @override
    def _convert_matches_into_route_keys(self, match: Match) -> Dict[str, str]:
        # In Bedrock Agents, all the parameters come inside the "parameters" key, not on the apiPath
        # So we have to search for route parameters in the parameters key
        parameters: Dict[str, str] = {}
        if match.groupdict() and self.current_event.parameters:
            parameters = {parameter["name"]: parameter["value"] for parameter in self.current_event.parameters}
        return parameters



import json
from typing import Any

from aws_lambda_powertools.utilities._data_masking.constants import DATA_MASKING_STRING


class BaseProvider:
    """
    When you try to create an instance of a subclass that does not implement the encrypt method,
    you will get a NotImplementedError with a message that says the method is not implemented:
    """

    def __init__(self, json_serializer=None, json_deserializer=None) -> None:
        self.json_serializer = json_serializer or self.default_json_serializer
        self.json_deserializer = json_deserializer or self.default_json_deserializer

    def default_json_serializer(self, data):
        return json.dumps(data).encode("utf-8")

    def default_json_deserializer(self, data):
        return json.loads(data.decode("utf-8"))

    def encrypt(self, data) -> str:
        raise NotImplementedError("Subclasses must implement encrypt()")

    def decrypt(self, data) -> Any:
        raise NotImplementedError("Subclasses must implement decrypt()")

    def mask(self, data) -> Any:
        if isinstance(data, (str, dict, bytes)):
            return DATA_MASKING_STRING
        elif isinstance(data, (list, tuple, set)):
            return type(data)([DATA_MASKING_STRING] * len(data))
        return DATA_MASKING_STRING



from __future__ import annotations

from typing import Any, Dict, List, Optional

from aws_lambda_powertools.utilities.data_classes.common import DictWrapper


def get_invoke_event(
    invoking_event: dict,
) -> AWSConfigConfigurationChanged | AWSConfigScheduledNotification | AWSConfigOversizedConfiguration:
    """
    Returns the corresponding event object based on the messageType in the invoking event.

    Parameters
    ----------
    invoking_event: dict
        The invoking event received.

    Returns
    -------
    AWSConfigConfigurationChanged | AWSConfigScheduledNotification | AWSConfigOversizedConfiguration:
        The event object based on the messageType in the invoking event.
    """

    message_type = invoking_event.get("messageType")

    if message_type == "ScheduledNotification":
        return AWSConfigScheduledNotification(invoking_event)

    if message_type == "OversizedConfigurationItemChangeNotification":
        return AWSConfigOversizedConfiguration(invoking_event)

    # Default return is AWSConfigConfigurationChanged event
    return AWSConfigConfigurationChanged(invoking_event)


class AWSConfigConfigurationChanged(DictWrapper):
    @property
    def configuration_item_diff(self) -> Dict:
        """The configuration item diff of the ConfigurationItemChangeNotification event."""
        return self["configurationItemDiff"]

    @property
    def configuration_item(self) -> AWSConfigConfigurationItemChanged:
        """The configuration item of the ConfigurationItemChangeNotification event."""
        return AWSConfigConfigurationItemChanged(self["configurationItem"])

    @property
    def raw_configuration_item(self) -> Dict:
        """The raw configuration item of the ConfigurationItemChangeNotification event."""
        return self["configurationItem"]

    @property
    def record_version(self) -> str:
        """The record version of the ConfigurationItemChangeNotification event."""
        return self["recordVersion"]

    @property
    def message_type(self) -> str:
        """The message type of the ConfigurationItemChangeNotification event."""
        return self["messageType"]

    @property
    def notification_creation_time(self) -> str:
        """The notification creation time of the ConfigurationItemChangeNotification event."""
        return self["notificationCreationTime"]


class AWSConfigConfigurationItemChanged(DictWrapper):
    @property
    def related_events(self) -> List:
        """The related events of the ConfigurationItemChangeNotification event."""
        return self["relatedEvents"]

    @property
    def relationships(self) -> List:
        """The relationships of the ConfigurationItemChangeNotification event."""
        return self["relationships"]

    @property
    def configuration(self) -> Dict:
        """The configuration of the ConfigurationItemChangeNotification event."""
        return self["configuration"]

    @property
    def supplementary_configuration(self) -> Dict:
        """The supplementary configuration of the ConfigurationItemChangeNotification event."""
        return self["supplementaryConfiguration"]

    @property
    def tags(self) -> Dict:
        """The tags of the ConfigurationItemChangeNotification event."""
        return self["tags"]

    @property
    def configuration_item_version(self) -> str:
        """The configuration item version of the ConfigurationItemChangeNotification event."""
        return self["configurationItemVersion"]

    @property
    def configuration_item_capture_time(self) -> str:
        """The configuration item capture time of the ConfigurationItemChangeNotification event."""
        return self["configurationItemCaptureTime"]

    @property
    def configuration_state_id(self) -> str:
        """The configuration state id of the ConfigurationItemChangeNotification event."""
        return self["configurationStateId"]

    @property
    def accountid(self) -> str:
        """The accountid of the ConfigurationItemChangeNotification event."""
        return self["awsAccountId"]

    @property
    def configuration_item_status(self) -> str:
        """The configuration item status of the ConfigurationItemChangeNotification event."""
        return self["configurationItemStatus"]

    @property
    def resource_type(self) -> str:
        """The resource type of the ConfigurationItemChangeNotification event."""
        return self["resourceType"]

    @property
    def resource_id(self) -> str:
        """The resource id of the ConfigurationItemChangeNotification event."""
        return self["resourceId"]

    @property
    def resource_name(self) -> str:
        """The resource name of the ConfigurationItemChangeNotification event."""
        return self["resourceName"]

    @property
    def resource_arn(self) -> str:
        """The resource arn of the ConfigurationItemChangeNotification event."""
        return self["ARN"]

    @property
    def region(self) -> str:
        """The region of the ConfigurationItemChangeNotification event."""
        return self["awsRegion"]

    @property
    def availability_zone(self) -> str:
        """The availability zone of the ConfigurationItemChangeNotification event."""
        return self["availabilityZone"]

    @property
    def configuration_state_md5_hash(self) -> str:
        """The md5 hash of the state of the ConfigurationItemChangeNotification event."""
        return self["configurationStateMd5Hash"]

    @property
    def resource_creation_time(self) -> str:
        """The resource creation time of the ConfigurationItemChangeNotification event."""
        return self["resourceCreationTime"]


class AWSConfigScheduledNotification(DictWrapper):
    @property
    def accountid(self) -> str:
        """The accountid of the ScheduledNotification event."""
        return self["awsAccountId"]

    @property
    def notification_creation_time(self) -> str:
        """The notification creation time of the ScheduledNotification event."""
        return self["notificationCreationTime"]

    @property
    def record_version(self) -> str:
        """The record version of the ScheduledNotification event."""
        return self["recordVersion"]

    @property
    def message_type(self) -> str:
        """The message type of the ScheduledNotification event."""
        return self["messageType"]


class AWSConfigOversizedConfiguration(DictWrapper):
    @property
    def configuration_item_summary(self) -> AWSConfigOversizedConfigurationItemSummary:
        """The configuration item summary of the OversizedConfiguration event."""
        return AWSConfigOversizedConfigurationItemSummary(self["configurationItemSummary"])

    @property
    def raw_configuration_item_summary(self) -> str:
        """The raw configuration item summary of the OversizedConfiguration event."""
        return self["configurationItemSummary"]

    @property
    def message_type(self) -> str:
        """The message type of the OversizedConfiguration event."""
        return self["messageType"]

    @property
    def notification_creation_time(self) -> str:
        """The notification creation time of the OversizedConfiguration event."""
        return self["notificationCreationTime"]

    @property
    def record_version(self) -> str:
        """The record version of the OversizedConfiguration event."""
        return self["recordVersion"]


class AWSConfigOversizedConfigurationItemSummary(DictWrapper):
    @property
    def change_type(self) -> str:
        """The change type of the OversizedConfiguration event."""
        return self["changeType"]

    @property
    def configuration_item_version(self) -> str:
        """The configuration item version of the OversizedConfiguration event."""
        return self["configurationItemVersion"]

    @property
    def configuration_item_capture_time(self) -> str:
        """The configuration item capture time of the OversizedConfiguration event."""
        return self["configurationItemCaptureTime"]

    @property
    def configuration_state_id(self) -> str:
        """The configuration state id of the OversizedConfiguration event."""
        return self["configurationStateId"]

    @property
    def accountid(self) -> str:
        """The accountid of the OversizedConfiguration event."""
        return self["awsAccountId"]

    @property
    def configuration_item_status(self) -> str:
        """The configuration item status of the OversizedConfiguration event."""
        return self["configurationItemStatus"]

    @property
    def resource_type(self) -> str:
        """The resource type of the OversizedConfiguration event."""
        return self["resourceType"]

    @property
    def resource_id(self) -> str:
        """The resource id of the OversizedConfiguration event."""
        return self["resourceId"]

    @property
    def resource_name(self) -> str:
        """The resource name of the OversizedConfiguration event."""
        return self["resourceName"]

    @property
    def resource_arn(self) -> str:
        """The resource arn of the OversizedConfiguration event."""
        return self["ARN"]

    @property
    def region(self) -> str:
        """The region of the OversizedConfiguration event."""
        return self["awsRegion"]

    @property
    def availability_zone(self) -> str:
        """The availability zone of the OversizedConfiguration event."""
        return self["availabilityZone"]

    @property
    def configuration_state_md5_hash(self) -> str:
        """The state md5 hash  of the OversizedConfiguration event."""
        return self["configurationStateMd5Hash"]

    @property
    def resource_creation_time(self) -> str:
        """The resource creation time of the OversizedConfiguration event."""
        return self["resourceCreationTime"]


class AWSConfigRuleEvent(DictWrapper):
    """Events for AWS Config Rules
    Documentation:
    --------------
    - https://docs.aws.amazon.com/config/latest/developerguide/evaluate-config_develop-rules_lambda-functions.html
    """

    def __init__(self, data: Dict[str, Any]):
        super().__init__(data)
        self._invoking_event: Optional[Any] = None
        self._rule_parameters: Optional[Any] = None

    @property
    def version(self) -> str:
        """The version of the event."""
        return self["version"]

    @property
    def invoking_event(
        self,
    ) -> AWSConfigConfigurationChanged | AWSConfigScheduledNotification | AWSConfigOversizedConfiguration:
        """The invoking payload of the event."""
        if self._invoking_event is None:
            self._invoking_event = self._json_deserializer(self["invokingEvent"])

        return get_invoke_event(self._invoking_event)

    @property
    def raw_invoking_event(self) -> str:
        """The raw invoking payload of the event."""
        return self["invokingEvent"]

    @property
    def rule_parameters(self) -> Dict:
        """The parameters of the event."""
        if self._rule_parameters is None:
            self._rule_parameters = self._json_deserializer(self["ruleParameters"])

        return self._rule_parameters

    @property
    def result_token(self) -> str:
        """The result token of the event."""
        return self["resultToken"]

    @property
    def event_left_scope(self) -> bool:
        """The left scope of the event."""
        return self["eventLeftScope"]

    @property
    def execution_role_arn(self) -> str:
        """The execution role arn of the event."""
        return self["executionRoleArn"]

    @property
    def config_rule_arn(self) -> str:
        """The arn of the rule of the event."""
        return self["configRuleArn"]

    @property
    def config_rule_name(self) -> str:
        """The name of the rule of the event."""
        return self["configRuleName"]

    @property
    def config_rule_id(self) -> str:
        """The id of the rule of the event."""
        return self["configRuleId"]

    @property
    def accountid(self) -> str:
        """The accountid of the event."""
        return self["accountId"]

    @property
    def evalution_mode(self) -> Optional[str]:
        """The evalution mode of the event."""
        return self.get("evaluationMode")



from typing import Dict, List, Optional

from pydantic import BaseModel, Field


class BedrockAgentModel(BaseModel):
    name: str
    id_: str = Field(..., alias="id")
    alias: str
    version: str


class BedrockAgentPropertyModel(BaseModel):
    name: str
    type_: str = Field(..., alias="type")
    value: str


class BedrockAgentRequestMediaModel(BaseModel):
    properties: List[BedrockAgentPropertyModel]


class BedrockAgentRequestBodyModel(BaseModel):
    content: Dict[str, BedrockAgentRequestMediaModel]


class BedrockAgentEventModel(BaseModel):
    message_version: str = Field(..., alias="messageVersion")
    input_text: str = Field(..., alias="inputText")
    session_id: str = Field(..., alias="sessionId")
    action_group: str = Field(..., alias="actionGroup")
    api_path: str = Field(..., alias="apiPath")
    http_method: str = Field(..., alias="httpMethod")
    session_attributes: Dict[str, str] = Field({}, alias="sessionAttributes")
    prompt_session_attributes: Dict[str, str] = Field({}, alias="promptSessionAttributes")
    agent: BedrockAgentModel
    parameters: Optional[List[BedrockAgentPropertyModel]] = None
    request_body: Optional[BedrockAgentRequestBodyModel] = Field(None, alias="requestBody")



"""
Datapane Reports Object

Describes the `Report` object
"""
from __future__ import annotations

import typing as t

from .dp_object import DPObjectRef


class CloudReport(DPObjectRef):
    """
    A Report that has been uploaded to datapane.com
    """

    # NOTE - uploading handled via the DPCloudUploader processor
    list_fields: t.List[str] = ["name", "web_url", "project"]
    endpoint: str = "/reports/"



from __future__ import annotations

from datapane.client import display_msg


class NotebookException(Exception):
    """Exception raised when a Notebook to Datapane conversion fails."""

    def _render_traceback_(self):
        display_msg(
            f"""**Conversion failed**

{str(self)}"""
        )


class NotebookParityException(NotebookException):
    """Exception raised when IPython output cache is not in sync with the saved notebook"""


class BlocksNotFoundException(NotebookException):
    """Exception raised when no blocks are found during conversion"""



import enum
import pathlib
import sys
import typing as t
from shlex import quote

from packaging.specifiers import SpecifierSet
from packaging.version import InvalidVersion


@enum.unique
class SupportedInstallers(str, enum.Enum):
    pip = "pip"

    def __str__(self):
        return self._name_

    @classmethod
    def as_list(cls):
        return [m._name_ for m in cls.__members__.values()]


DEFAULT_INSTALLER = SupportedInstallers.pip
DEFAULT_PYTHON_VERSION = "{version.major}.{version.minor}".format(version=sys.version_info)
SUPPORTED_PYTHON_VERSIONS = SpecifierSet(">=3.8,<3.12")


# use printf-style formatting, to avoid conflicting with sh/docker variables
DOCKERFILE_TEMPLATE = r"""
ARG PYTHON_VERSION='%(python_version)s'

FROM python:${PYTHON_VERSION}
WORKDIR /app/

%(install_deps_block)s

# load your files into the image
# To exclude files, setup a .dockerignore
COPY ./ .

%(notebook_convert_block)s

# execute your app (what to run on the server)
CMD exec python %(app_file)s
""".lstrip()

_pip_install_block = r"""
# Install the App dependencies
# Note: If your requirements.txt file depends on other files,
#       you should change to 'COPY ./ .'.
# If you don't have a requirements.txt file:
#  - remove the below lines
#  - let us know your usecase, so we can improve Datapane!
COPY ./requirements.txt .
RUN \
  echo "Installing dependencies from requirements.txt..." \
  && pip install --no-cache-dir -r requirements.txt
""".strip()

_notebook_convert_block = r"""
# Note: Datapane currently expects a python file to execute.
#       This converts your notebook to a python file.
# If you face issues, please let us know!
RUN \
  echo "Generating %(output_path)s from your notebook %(notebook_path)s..." \
  && datapane app generate script-from-nb --nb %(notebook_path)s --out %(output_path)s
"""


def generate_dockerfile(
    python_version: str = DEFAULT_PYTHON_VERSION,
    installer: SupportedInstallers = DEFAULT_INSTALLER,
    app_file: t.Union[str, pathlib.Path, None] = None,
):
    # TODO: better error handling
    assert _is_valid_python_version(python_version), "The python version you specified is not supported"
    assert _is_valid_installer(installer), "The chosen installer is not supported"

    if installer == SupportedInstallers.pip:
        install_deps_block = _pip_install_block
    else:
        # TODO: we shouldn't ever get here
        install_deps_block = ""

    if app_file is None:
        app_file = "<app>.py"
    app_file = pathlib.Path(app_file)

    if app_file.suffix == ".ipynb":
        notebook_convert_block = _notebook_convert_block % dict(
            notebook_path=q(app_file),
            output_path=q(app_file.with_suffix(".py")),
        )
        app_file = app_file.with_suffix(".py")
    else:
        notebook_convert_block = ""

    return DOCKERFILE_TEMPLATE % dict(
        python_version=python_version,
        install_deps_block=install_deps_block,
        notebook_convert_block=notebook_convert_block,
        app_file=q(app_file),
    )


def python_files(cwd: t.Union[str, pathlib.Path] = ".", /, extensions: t.Sequence[str] = (".py", ".ipynb")):
    cwd = pathlib.Path(cwd) if isinstance(cwd, str) else cwd

    # iterdir() doesn't define an order.
    # 'sort' it for consistency on ipynb vs py duplicates.
    for path in sorted(cwd.iterdir()):
        if path.suffix not in extensions:
            continue
        # ignoring "hidden" files - POSIX and Python style
        if path.name.startswith((".", "_")):
            continue
        if not path.is_file():
            continue
        yield path


def _is_valid_python_version(python_version: str) -> bool:
    try:
        return SUPPORTED_PYTHON_VERSIONS.contains(python_version)
    except InvalidVersion:
        return False


def _is_valid_installer(installer: t.Union[SupportedInstallers, str]) -> bool:
    return installer in SupportedInstallers.__members__


def q(s: t.Union[str, pathlib.Path]) -> str:
    """Shell-quote a string or path."""
    return quote(str(s))



"""This module provides a simple interface to capture analytics data"""
import atexit
import os
import platform
import queue
import sys
import threading
from contextlib import suppress
from functools import wraps
from pathlib import Path
from time import sleep, time
from typing import Any, Callable, Dict, Optional, TypeVar, cast

import posthog

from . import config as c
from .utils import IN_PYTEST, ON_DATAPANE, log

posthog.sync_mode = True
posthog.api_key = "phc_wxtD2Qxd3RMlmCCSYDC0rW1We22yh06cMcffnfSJTZy"
posthog.host = "https://events.datapane.com/"
_NO_ANALYTICS_FILE: Path = c.APP_DIR / "no_analytics"
MAX_QUEUE_SIZE = 250
QUEUE_TIMEOUT = 10
IN_DPSERVER = "dp" in sys.modules
USING_CONDA = os.path.exists(os.path.join(sys.prefix, "conda-meta", "history"))


class Consumer(threading.Thread):
    """Maintains and consumes queue for posthog requests"""

    def __init__(self, *a, **kw):
        super().__init__(*a, **kw, name="Analytics event consumer", daemon=True)
        self.queue = queue.Queue(MAX_QUEUE_SIZE)

    def run(self):
        """
        Run the next function at the start of the queue whenever the queue is non-empty.
        Note `self.queue.get()` is blocking while the queue is empty
        """
        while True:
            (f, args) = self.queue.get()
            with suppress(Exception):
                f(*args)
            self.queue.task_done()

    def send(self, f: Callable[..., None], args: Any):
        """Add analytics function and arguments to the thread queue"""
        self.queue.put((f, args))

    def join_queue(self):
        """
        Poll `queue.unfinished_tasks` until a max time is reached,
        rather than calling `queue.join`, as `queue.Queue.join` has no `timeout` parameter
        """
        end_time = time() + QUEUE_TIMEOUT
        while self.queue.unfinished_tasks and time() < end_time:
            sleep(1)


def is_analytics_disabled() -> bool:
    """Determine the initial state for analytics if not already set"""
    # disable if globally disabled or in certain envs
    if _NO_ANALYTICS_FILE.exists() or ON_DATAPANE or IN_PYTEST or IN_DPSERVER:
        log.debug("Analytics disabled")
        return True
    return False


_NO_ANALYTICS: bool = is_analytics_disabled()


def user_properties() -> Dict:
    """User properties to be added on `identify`, and amended on `capture`"""
    from datapane import __version__
    from datapane.ipython.environment import get_environment

    environment = get_environment()

    return dict(
        os=platform.system(),
        python_version=platform.python_version(),
        dp_version=__version__,
        environment_type=environment.name,
        in_jupyter=environment.is_notebook_environment,
        using_conda=USING_CONDA,
    )


_consumer: Optional[Consumer] = None


def start_queue():
    global _consumer
    _consumer = Consumer()
    _consumer.start()


if not _NO_ANALYTICS:
    start_queue()


def capture(event: str, config: Optional[c.Config] = None, **properties) -> None:
    # Used for capturing generic events with properties
    if _NO_ANALYTICS:
        return None
    config = config or c.get_config()
    user_props = user_properties()

    # Note "$" isn't a valid char in `dict(k=v)` keys
    properties.update({"source": "cli", **user_props, "$set": user_props.copy()})

    _consumer.send(posthog.capture, args=[config.session_id, event, properties])


# def identify(config: c.Config, **properties) -> None:
#     properties.update(user_properties())
#     _consumer.send(posthog.identify, args=[config.session_id, properties])


# def capture_init(config: c.Config) -> None:
#     # Generates an identify event on init
#     if _NO_ANALYTICS:
#         return None
#     identify(config.session_id)

# From the mypy docs ...https://mypy.readthedocs.io/en/stable/generics.html#declaring-decorators
F = TypeVar("F", bound=Callable[..., Any])


def capture_event(name: str) -> Callable[[F], F]:
    """Decorator to capture 'name' as analytics event when the function is called"""

    def _decorator(func: F) -> F:
        @wraps(func)
        def _wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            finally:
                capture(name)

        return cast(F, _wrapper)

    return _decorator


@atexit.register
def finish_queue():
    """Block main thread until analytics queue is clear"""
    global _consumer
    if _consumer is not None:
        _consumer.join_queue()
    _consumer = None



import altair as alt
from vega_datasets import data

import datapane as dp

df = data.iris()
fig = alt.Chart(df).mark_point().encode(x="petalLength:Q", y="petalWidth:Q", color="species:N")


report = dp.App(
    dp.Text(
        """
# Hello world

Welcome to [Datapane](https://datapane.com)! Built for data developers, Datapane makes it simple to create beautiful reports from anywhere we can run Python.

Reports can be exported as standalone HTML documents or hosted on Datapane Cloud, where they can be shared via a link or embedded.

This [_Hello world_](https://github.com/datapane/datapane/blob/master/src/datapane/resources/templates/hello/hello.py) page itself was generated with Datapane!
        """
    ),
    dp.Text(
        """
## Our first Report

Let’s create a simple report! In this scenario, we have a dataset and visualization that we need to share with our colleagues.
        """
    ),
    dp.Code(
        """
import altair as alt
from vega_datasets import data

df = data.iris()
fig = (
    alt.Chart(df)
    .mark_point()
    .encode(x="petalLength:Q", y="petalWidth:Q", color="species:N")
)
        """
    ),
    dp.Text(
        """
        Datapane makes generating a report simple – all it takes is the following.
        """
    ),
    dp.Code(
        """
import datapane as dp

report = dp.App(
    dp.Plot(fig),
    dp.DataTable(df)
)

report.save(path="my_report.html")
        """
    ),
    dp.Text(
        """
This will generate an HTML file, `my_report.html`, containing our interactive report:
        """
    ),
    dp.Plot(fig),
    dp.DataTable(df),
    dp.Text(
        """
We can now open our standalone HTML report with a web browser and email it to our colleagues who can do the same.

But what if we want to send a link to our report or embed it instead?

## Datapane Cloud makes it simple to share

Datapane Cloud is free! Share reports privately, or embed them into platforms like Salesforce and Medium.

To sign up go to https://cloud.datapane.com/accounts/signup/ and then login by doing the following
"""
    ),
    dp.Select(
        dp.Code("$ datapane login", label="Terminal"),
        dp.Code("!datapane login", label="Python/Jupyter"),
    ),
    dp.Text(
        """
Once we've logged in, [the API key](https://datapane.com/settings/) will be stored in the current environment.
        """
    ),
    dp.Code(
        """
report.upload(name="My report")
        """
    ),
    dp.Text(
        """
By using `upload()` in place of `save()`, our report is uploaded to Datapane Cloud and a shareable link is generated.

We can now send this link to our colleagues or use it to embed our report on another platform.

## Next steps

There's so much more we can do with Datapane!
- [Read the documentation](https://docs.datapane.com)
- [Explore our community spaces](https://datapane.com/community)
- [We're open-source – browse the Datapane repo](https://github.com/datapane/datapane)
- [Sign up for a Datapane Cloud account](https://cloud.datapane.com/accounts/signup/)
"""
    ),
)

report.save(
    path="hello.html",
    open=True,
)

# You can also upload your report to a Datapane Server by logging in then running the following
# report.upload(name="hello-world", open=True)



import json
from typing import List, Optional

from pydantic import BaseModel, PositiveInt, validator

from aws_lambda_powertools.shared.functions import base64_decode
from aws_lambda_powertools.utilities.parser.models import KinesisFirehoseRecordMetadata

from .sqs import SqsRecordModel


class KinesisFirehoseSqsRecord(BaseModel):
    data: SqsRecordModel
    recordId: str
    approximateArrivalTimestamp: PositiveInt
    kinesisRecordMetadata: Optional[KinesisFirehoseRecordMetadata] = None

    @validator("data", pre=True, allow_reuse=True)
    def data_base64_decode(cls, value):
        # Firehose payload is encoded
        return json.loads(base64_decode(value))


class KinesisFirehoseSqsModel(BaseModel):
    invocationId: str
    deliveryStreamArn: str
    region: str
    sourceKinesisStreamArn: Optional[str] = None
    records: List[KinesisFirehoseSqsRecord]



# Copyright 2020 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Data Commons Python API Stat Module.

Provides functions for getting data on StatisticalVariables from Data Commons Graph.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import six

import datacommons.utils as utils

# stat_var specific batch size.
_STAT_BATCH_SIZE = 2000


def get_stat_value(place,
                   stat_var,
                   date=None,
                   measurement_method=None,
                   observation_period=None,
                   unit=None,
                   scaling_factor=None):
    """Returns a value for `place` based on the `stat_var`.

    Args:
      place (`str`): The dcid of Place to query for.
      stat_var (`str`): The dcid of the StatisticalVariable.
      date (`str`): Optional, the preferred date of observation
        in ISO 8601 format. If not specified, returns the latest observation.
      measurement_method (`str`): Optional, the dcid of the preferred
        `measurementMethod` value.
      observation_period (`str`): Optional, the preferred
        `observationPeriod` value.
      unit (`str`): Optional, the dcid of the preferred `unit` value.
      scaling_factor (`int`): Optional, the preferred `scalingFactor` value.
    Returns:
      A `float` the value of `stat_var` for `place`, filtered
      by optional args. If no data, returns nan.

    Raises:
      ValueError: If the payload returned by the Data Commons REST API is
        malformed.

    Examples:
      >>> get_stat_value("geoId/05", "Count_Person")
          366331
    """
    url = utils._API_ROOT + utils._API_ENDPOINTS['get_stat_value']
    url += '?place={}&stat_var={}'.format(place, stat_var)
    if date:
        url += '&date={}'.format(date)
    if measurement_method:
        url += '&measurement_method={}'.format(measurement_method)
    if observation_period:
        url += '&observation_period={}'.format(observation_period)
    if unit:
        url += '&unit={}'.format(unit)
    if scaling_factor:
        url += '&scaling_factor={}'.format(scaling_factor)

    try:
        res_json = utils._send_request(url, post=False, use_payload=False)
    except ValueError:
        return float('nan')
    if 'value' not in res_json:
        return float('nan')
    return res_json['value']


def get_stat_series(place,
                    stat_var,
                    measurement_method=None,
                    observation_period=None,
                    unit=None,
                    scaling_factor=None):
    """Returns a `dict` mapping dates to value of `stat_var` for `place`.

    Args:
      place (`str`): The dcid of Place to query for.
      stat_var (`str`): The dcid of the StatisticalVariable.
      measurement_method (`str`): Optional, the dcid of the preferred
        `measurementMethod` value.
      observation_period (`str`): Optional, the preferred
        `observationPeriod` value.
      unit (`str`): Optional, the dcid of the preferred `unit` value.
      scaling_factor (`int`): Optional, the preferred `scalingFactor` value.
    Returns:
      A `dict` mapping dates to value of `stat_var` for `place`,
      representing a time series that satisfies all input parameters.

    Raises:
      ValueError: If the payload returned by the Data Commons REST API is
        malformed.

    Examples:
      >>> get_stat_series("geoId/05", "Count_Person")
          {"1962":17072000,"2009":36887615,"1929":5531000,"1930":5711000}
    """
    url = utils._API_ROOT + utils._API_ENDPOINTS['get_stat_series']
    url += '?place={}&stat_var={}'.format(place, stat_var)
    if measurement_method:
        url += '&measurement_method={}'.format(measurement_method)
    if observation_period:
        url += '&observation_period={}'.format(observation_period)
    if unit:
        url += '&unit={}'.format(unit)
    if scaling_factor:
        url += '&scaling_factor={}'.format(scaling_factor)
    
    try:
        res_json = utils._send_request(url, post=False, use_payload=False)
    except ValueError:
        return {}

    if 'series' not in res_json:
        return {}
    return res_json['series']


def get_stat_all(places, stat_vars):
    """Returns a nested `dict` of all time series for `places` and `stat_vars`.

    Args:
      places (`Iterable` of `str`): The dcids of Places to query for.
      stat_vars (`Iterable` of `str`): The dcids of the StatisticalVariables.
    Returns:
      A nested `dict` mapping Places to StatisticalVariables and all available
      time series for each Place and StatisticalVariable pair.

    Raises:
      ValueError: If the payload returned by the Data Commons REST API is
        malformed.

    Examples:
      >>> get_stat_all(["geoId/05", "geoId/06"], ["Count_Person", "Count_Person_Male"])
      {
        "geoId/05": {
          "Count_Person": {
            "sourceSeries": [
              {
                "val": {
                  "2010": 1633,
                  "2011": 1509,
                  "2012": 1581,
                },
                "observationPeriod": "P1Y",
                "importName": "Wikidata",
                "provenanceDomain": "wikidata.org"
              },
              {
                "val": {
                  "2010": 1333,
                  "2011": 1309,
                  "2012": 131,
                },
                "observationPeriod": "P1Y",
                "importName": "CensusPEPSurvey",
                "provenanceDomain": "census.gov"
              }
            ],
            }
          },
          "Count_Person_Male": {
            "sourceSeries": [
              {
                "val": {
                  "2010": 1633,
                  "2011": 1509,
                  "2012": 1581,
                },
                "observationPeriod": "P1Y",
                "importName": "CensusPEPSurvey",
                "provenanceDomain": "census.gov"
              }
            ],
          }
        },
        "geoId/02": {
          "Count_Person": {},
          "Count_Person_Male": {
              "sourceSeries": [
                {
                  "val": {
                    "2010": 13,
                    "2011": 13,
                    "2012": 322,
                  },
                  "observationPeriod": "P1Y",
                  "importName": "CensusPEPSurvey",
                  "provenanceDomain": "census.gov"
                }
              ]
            }
        }
      }
    """
    url = utils._API_ROOT + utils._API_ENDPOINTS['get_stat_all']
    # Cast iterable-like to list.
    places = list(places)
    stat_vars = list(stat_vars)

    # Aiming for _STAT_BATCH_SIZE entries total.
    # _STAT_BATCH_SIZE = num places x num stat_vars, so aim for
    # _STAT_BATCH_SIZE/len(stat_vars) places per batch.
    places_per_batch = _STAT_BATCH_SIZE // len(stat_vars)
    # Get number of batches via an arithmetic ceiling trick:
    # 11//10 rounds down to 1.
    # -11//10 rounds down to -2.
    # We can divide with, then remove the negative to get the ceiling.
    batches = -(-len(places) // places_per_batch)
    res = {}
    for i in range(batches):
        req_json = {
            'stat_vars': stat_vars,
            'places': places[i * places_per_batch:(i + 1) * places_per_batch]
        }
        # Send the request
        res_json = utils._send_request(url,
                                       req_json=req_json,
                                       use_payload=False)
        if 'placeData' not in res_json:
            # The REST API spec will always return a dictionary under
            # placeData, even if no places exist or have no
            # data. If no Places are provided, REST will return an
            # error, which will have been caught and passed on in
            # _send_request.
            raise ValueError("Unexpected response from REST stat/all API.")

        # Unnest the REST response for keys that have single-element values.
        place_statvar_series = collections.defaultdict(dict)
        for place_dcid, place in res_json['placeData'].items():
            stat_var_data = place.get('statVarData')
            if not stat_var_data:
                # The REST API spec will always return a dictionary under
                # statVarData, even if no StatVars exist or have no
                # data. If no StatVars are provided, REST will return an
                # error, which will have been caught and passed on in
                # _send_request.
                raise ValueError("Unexpected response from REST stat/all API.")
            for stat_var_dcid, stat_var in stat_var_data.items():
                place_statvar_series[place_dcid][stat_var_dcid] = stat_var
        res.update(dict(place_statvar_series))

    return res



from typing import Any, Dict, Type

from pydantic import BaseModel

from aws_lambda_powertools.utilities.idempotency.exceptions import (
    IdempotencyModelTypeError,
    IdempotencyNoSerializationModelError,
)
from aws_lambda_powertools.utilities.idempotency.serialization.base import (
    BaseIdempotencyModelSerializer,
    BaseIdempotencySerializer,
)


class PydanticSerializer(BaseIdempotencyModelSerializer):
    """Pydantic serializer for idempotency models"""

    def __init__(self, model: Type[BaseModel]):
        """
        Parameters
        ----------
        model: Model
            Pydantic model to be used for serialization
        """
        self.__model: Type[BaseModel] = model

    def to_dict(self, data: BaseModel) -> Dict:
        if callable(getattr(data, "model_dump", None)):
            # Support for pydantic V2
            return data.model_dump()  # type: ignore[unused-ignore,attr-defined]
        return data.dict()

    def from_dict(self, data: Dict) -> BaseModel:
        if callable(getattr(self.__model, "model_validate", None)):
            # Support for pydantic V2
            return self.__model.model_validate(data)  # type: ignore[unused-ignore,attr-defined]
        return self.__model.parse_obj(data)

    @classmethod
    def instantiate(cls, model_type: Any) -> BaseIdempotencySerializer:
        if model_type is None:
            raise IdempotencyNoSerializationModelError("No serialization model was supplied")

        if not issubclass(model_type, BaseModel):
            raise IdempotencyModelTypeError("Model type is not inherited from pydantic BaseModel")

        return cls(model=model_type)



from abc import ABC, abstractmethod


class IReporting(ABC):
    def __init__(self):
        pass

    @abstractmethod
    def link_report(
        self,
        polly_session,
        repo_key: str,
        dataset_id: str,
        workspace_id: int,
        workspace_path: str,
        access_key: str,
    ):
        pass

    @abstractmethod
    def link_report_url(
        self, polly_session, repo_key: str, dataset_id: str, url_to_be_linked: str
    ):
        pass

    @abstractmethod
    def fetch_linked_reports(self, polly_session, repo_key: str, dataset_id: str):
        pass

    @abstractmethod
    def delete_linked_report(
        self, polly_session, repo_key: str, dataset_id: str, report_id: str
    ):
        pass



"""By default, glom aims to safely return a transformed copy of your
data. But sometimes you really need to transform an existing object.

When you already have a large or complex bit of nested data that you
are sure you want to modify in-place, glom has you covered, with the
:func:`~glom.assign` function, and the :func:`~glom.Assign` specifier
type.

"""
import operator
from pprint import pprint

from .core import Path, T, S, Spec, glom, UnregisteredTarget, GlomError, PathAccessError, UP
from .core import TType, register_op, TargetRegistry, bbrepr, PathAssignError, arg_val, _assign_op


try:
    basestring
except NameError:
    basestring = str


if getattr(__builtins__, '__dict__', None) is not None:
    # pypy's __builtins__ is a module, as is CPython's REPL, but at
    # normal execution time it's a dict?
    __builtins__ = __builtins__.__dict__


class PathDeleteError(PathAssignError):
    """This :exc:`GlomError` subtype is raised when an assignment fails,
    stemming from an :func:`~glom.delete` call or other
    :class:`~glom.Delete` usage.

    One example would be deleting an out-of-range position in a list::

      >>> delete(["short", "list"], Path(5))
      Traceback (most recent call last):
      ...
      PathDeleteError: could not delete 5 on object at Path(), got error: IndexError(...

    Other assignment failures could be due to deleting a read-only
    ``@property`` or exception being raised inside a ``__delattr__()``.

    """
    def get_message(self):
        return ('could not delete %r on object at %r, got error: %r'
                % (self.dest_name, self.path, self.exc))


def _apply_for_each(func, path, val):
    layers = path.path_t.__stars__()
    if layers:
        for i in range(layers - 1):
            val = sum(val, [])  # flatten out the extra layers
        for inner in val:
            func(inner)
    else:
        func(val)


class Assign(object):
    """*New in glom 18.3.0*

    The ``Assign`` specifier type enables glom to modify the target,
    performing a "deep-set" to mirror glom's original deep-get use
    case.

    ``Assign`` can be used to perform spot modifications of large data
    structures when making a copy is not desired::

      # deep assignment into a nested dictionary
      >>> target = {'a': {}}
      >>> spec = Assign('a.b', 'value')
      >>> _ = glom(target, spec)
      >>> pprint(target)
      {'a': {'b': 'value'}}

    The value to be assigned can also be a :class:`~glom.Spec`, which
    is useful for copying values around within the data structure::

      # copying one nested value to another
      >>> _ = glom(target, Assign('a.c', Spec('a.b')))
      >>> pprint(target)
      {'a': {'b': 'value', 'c': 'value'}}

    Another handy use of Assign is to deep-apply a function::

      # sort a deep nested list
      >>> target={'a':{'b':[3,1,2]}}
      >>> _ = glom(target, Assign('a.b', Spec(('a.b',sorted))))
      >>> pprint(target)
      {'a': {'b': [1, 2, 3]}}

    Like many other specifier types, ``Assign``'s destination path can be
    a :data:`~glom.T` expression, for maximum control::

      # changing the error message of an exception in an error list
      >>> err = ValueError('initial message')
      >>> target = {'errors': [err]}
      >>> _ = glom(target, Assign(T['errors'][0].args, ('new message',)))
      >>> str(err)
      'new message'

    ``Assign`` has built-in support for assigning to attributes of
    objects, keys of mappings (like dicts), and indexes of sequences
    (like lists). Additional types can be registered through
    :func:`~glom.register()` using the ``"assign"`` operation name.

    Attempting to assign to an immutable structure, like a
    :class:`tuple`, will result in a
    :class:`~glom.PathAssignError`. Attempting to assign to a path
    that doesn't exist will raise a :class:`~PathAccessError`.

    To automatically backfill missing structures, you can pass a
    callable to the *missing* argument. This callable will be called
    for each path segment along the assignment which is not
    present.

       >>> target = {}
       >>> assign(target, 'a.b.c', 'hi', missing=dict)
       {'a': {'b': {'c': 'hi'}}}

    """
    def __init__(self, path, val, missing=None):
        # TODO: an option like require_preexisting or something to
        # ensure that a value is mutated, not just added. Current
        # workaround is to do a Check().
        if isinstance(path, basestring):
            path = Path.from_text(path)
        elif type(path) is TType:
            path = Path(path)
        elif not isinstance(path, Path):
            raise TypeError('path argument must be a .-delimited string, Path, T, or S')

        try:
            self.op, self.arg = path.items()[-1]
        except IndexError:
            raise ValueError('path must have at least one element')
        self._orig_path = path
        self.path = path[:-1]

        if self.op not in '[.P':
            # maybe if we add null-coalescing this should do something?
            raise ValueError('last part of path must be setattr or setitem')
        self.val = val

        if missing is not None:
            if not callable(missing):
                raise TypeError('expected missing to be callable, not %r' % (missing,))
        self.missing = missing

    def glomit(self, target, scope):
        val = arg_val(target, self.val, scope)

        op, arg, path = self.op, self.arg, self.path
        if self.path.startswith(S):
            dest_target = scope[UP]
            dest_path = self.path.from_t()
        else:
            dest_target = target
            dest_path = self.path
        try:
            dest = scope[glom](dest_target, dest_path, scope)
        except PathAccessError as pae:
            if not self.missing:
                raise

            remaining_path = self._orig_path[pae.part_idx + 1:]
            val = scope[glom](self.missing(), Assign(remaining_path, val, missing=self.missing), scope)

            op, arg = self._orig_path.items()[pae.part_idx]
            path = self._orig_path[:pae.part_idx]
            dest = scope[glom](dest_target, path, scope)

        # TODO: forward-detect immutable dest?
        _apply = lambda dest: _assign_op(
            dest=dest, op=op, arg=arg, val=val, path=path, scope=scope)
        _apply_for_each(_apply, path, dest)

        return target

    def __repr__(self):
        cn = self.__class__.__name__
        if self.missing is None:
            return '%s(%r, %r)' % (cn, self._orig_path, self.val)
        return '%s(%r, %r, missing=%s)' % (cn, self._orig_path, self.val, bbrepr(self.missing))


def assign(obj, path, val, missing=None):
    """*New in glom 18.3.0*

    The ``assign()`` function provides convenient "deep set"
    functionality, modifying nested data structures in-place::

      >>> target = {'a': [{'b': 'c'}, {'d': None}]}
      >>> _ = assign(target, 'a.1.d', 'e')  # let's give 'd' a value of 'e'
      >>> pprint(target)
      {'a': [{'b': 'c'}, {'d': 'e'}]}

    Missing structures can also be automatically created with the
    *missing* parameter. For more information and examples, see the
    :class:`~glom.Assign` specifier type, which this function wraps.
    """
    return glom(obj, Assign(path, val, missing=missing))


_ALL_BUILTIN_TYPES = [v for v in __builtins__.values() if isinstance(v, type)]
_BUILTIN_BASE_TYPES = [v for v in _ALL_BUILTIN_TYPES
                       if not issubclass(v, tuple([t for t in _ALL_BUILTIN_TYPES
                                                   if t not in (v, type, object)]))]
_UNASSIGNABLE_BASE_TYPES = tuple(set(_BUILTIN_BASE_TYPES)
                                 - set([dict, list, BaseException, object, type]))


def _set_sequence_item(target, idx, val):
    target[int(idx)] = val


def _assign_autodiscover(type_obj):
    # TODO: issubclass or "in"?
    if issubclass(type_obj, _UNASSIGNABLE_BASE_TYPES):
        return False

    if callable(getattr(type_obj, '__setitem__', None)):
        if callable(getattr(type_obj, 'index', None)):
            return _set_sequence_item
        return operator.setitem

    return setattr


register_op('assign', auto_func=_assign_autodiscover, exact=False)


class Delete(object):
    """
    In addition to glom's core "deep-get" and ``Assign``'s "deep-set",
    the ``Delete`` specifier type performs a "deep-del", which can
    remove items from larger data structures by key, attribute, and
    index.

    >>> target = {'dict': {'x': [5, 6, 7]}}
    >>> glom(target, Delete('dict.x.1'))
    {'dict': {'x': [5, 7]}}
    >>> glom(target, Delete('dict.x'))
    {'dict': {}}

    If a target path is missing, a :exc:`PathDeleteError` will be
    raised. To ignore missing targets, use the ``ignore_missing``
    flag:

    >>> glom(target, Delete('does_not_exist', ignore_missing=True))
    {'dict': {}}

    ``Delete`` has built-in support for deleting attributes of
    objects, keys of dicts, and indexes of sequences
    (like lists). Additional types can be registered through
    :func:`~glom.register()` using the ``"delete"`` operation name.

    .. versionadded:: 20.5.0
    """
    def __init__(self, path, ignore_missing=False):
        if isinstance(path, basestring):
            path = Path.from_text(path)
        elif type(path) is TType:
            path = Path(path)
        elif not isinstance(path, Path):
            raise TypeError('path argument must be a .-delimited string, Path, T, or S')

        try:
            self.op, self.arg = path.items()[-1]
        except IndexError:
            raise ValueError('path must have at least one element')
        self._orig_path = path
        self.path = path[:-1]

        if self.op not in '[.P':
            raise ValueError('last part of path must be an attribute or index')

        self.ignore_missing = ignore_missing

    def _del_one(self, dest, op, arg, scope):
        if op == '[':
            try:
                del dest[arg]
            except IndexError as e:
                if not self.ignore_missing:
                    raise PathDeleteError(e, self.path, arg)
        elif op == '.':
            try:
                delattr(dest, arg)
            except AttributeError as e:
                if not self.ignore_missing:
                    raise PathDeleteError(e, self.path, arg)
        elif op == 'P':
            _delete = scope[TargetRegistry].get_handler('delete', dest)
            try:
                _delete(dest, arg)
            except Exception as e:
                if not self.ignore_missing:
                    raise PathDeleteError(e, self.path, arg)

    def glomit(self, target, scope):
        op, arg, path = self.op, self.arg, self.path
        if self.path.startswith(S):
            dest_target = scope[UP]
            dest_path = self.path.from_t()
        else:
            dest_target = target
            dest_path = self.path
        try:
            dest = scope[glom](dest_target, dest_path, scope)
        except PathAccessError as pae:
            if not self.ignore_missing:
                raise
        else:
            _apply_for_each(lambda dest: self._del_one(dest, op, arg, scope), path, dest)

        return target

    def __repr__(self):
        cn = self.__class__.__name__
        return '%s(%r)' % (cn, self._orig_path)


def delete(obj, path, ignore_missing=False):
    """
    The ``delete()`` function provides "deep del" functionality,
    modifying nested data structures in-place::

      >>> target = {'a': [{'b': 'c'}, {'d': None}]}
      >>> delete(target, 'a.0.b')
      {'a': [{}, {'d': None}]}

    Attempting to delete missing keys, attributes, and indexes will
    raise a :exc:`PathDeleteError`. To ignore these errors, use the
    *ignore_missing* argument::

      >>> delete(target, 'does_not_exist', ignore_missing=True)
      {'a': [{}, {'d': None}]}

    For more information and examples, see the :class:`~glom.Delete`
    specifier type, which this convenience function wraps.

    .. versionadded:: 20.5.0
    """
    return glom(obj, Delete(path, ignore_missing=ignore_missing))


def _del_sequence_item(target, idx):
    del target[int(idx)]


def _delete_autodiscover(type_obj):
    if issubclass(type_obj, _UNASSIGNABLE_BASE_TYPES):
        return False

    if callable(getattr(type_obj, '__delitem__', None)):
        if callable(getattr(type_obj, 'index', None)):
            return _del_sequence_item
        return operator.delitem
    return delattr


register_op('delete', auto_func=_delete_autodiscover, exact=False)



# NOTE: keeps for compatibility
from __future__ import annotations

from typing import Any, Dict, List

from aws_lambda_powertools.metrics.provider.datadog.datadog import DatadogProvider
from aws_lambda_powertools.shared.types import AnyCallableT


class DatadogMetrics:
    """
    DatadogProvider creates metrics asynchronously via Datadog extension or exporter.

    **Use `aws_lambda_powertools.DatadogMetrics` to create and metrics to Datadog.**

    Example
    -------
    **Creates a few metrics and publish at the end of a function execution**

        from aws_lambda_powertools.metrics.provider.datadog import DatadogMetrics

        metrics = DatadogMetrics(namespace="ServerlessAirline")

        @metrics.log_metrics(capture_cold_start_metric=True)
        def lambda_handler():
            metrics.add_metric(name="item_sold", value=1, product="latte", order="online")
            return True

    Environment variables
    ---------------------
    POWERTOOLS_METRICS_NAMESPACE : str
        metric namespace

    Parameters
    ----------
    flush_to_log : bool, optional
        Used when using export instead of Lambda Extension
    namespace : str, optional
        Namespace for metrics
    provider: DatadogProvider, optional
        Pre-configured DatadogProvider provider

    Raises
    ------
    MetricValueError
        When metric value isn't a number
    SchemaValidationError
        When metric object fails Datadog schema validation
    """

    # NOTE: We use class attrs to share metrics data across instances
    # this allows customers to initialize Metrics() throughout their code base (and middlewares)
    # and not get caught by accident with metrics data loss, or data deduplication
    # e.g., m1 and m2 add metric ProductCreated, however m1 has 'version' dimension  but m2 doesn't
    # Result: ProductCreated is created twice as we now have 2 different EMF blobs
    _metrics: List = []
    _default_tags: Dict[str, Any] = {}

    def __init__(
        self,
        namespace: str | None = None,
        flush_to_log: bool | None = None,
        provider: DatadogProvider | None = None,
    ):
        self.metric_set = self._metrics
        self.default_tags = self._default_tags

        if provider is None:
            self.provider = DatadogProvider(
                namespace=namespace,
                flush_to_log=flush_to_log,
                metric_set=self.metric_set,
            )
        else:
            self.provider = provider

    def add_metric(
        self,
        name: str,
        value: float,
        timestamp: int | None = None,
        **tags: Any,
    ) -> None:
        self.provider.add_metric(name=name, value=value, timestamp=timestamp, **tags)

    def serialize_metric_set(self, metrics: List | None = None) -> List:
        return self.provider.serialize_metric_set(metrics=metrics)

    def flush_metrics(self, raise_on_empty_metrics: bool = False) -> None:
        self.provider.flush_metrics(raise_on_empty_metrics=raise_on_empty_metrics)

    def log_metrics(
        self,
        lambda_handler: AnyCallableT | None = None,
        capture_cold_start_metric: bool = False,
        raise_on_empty_metrics: bool = False,
        default_tags: Dict[str, Any] | None = None,
    ):
        return self.provider.log_metrics(
            lambda_handler=lambda_handler,
            capture_cold_start_metric=capture_cold_start_metric,
            raise_on_empty_metrics=raise_on_empty_metrics,
            default_tags=default_tags,
        )

    def set_default_tags(self, **tags) -> None:
        self.provider.set_default_tags(**tags)
        self.default_tags.update(**tags)

    def clear_metrics(self) -> None:
        self.provider.clear_metrics()

    def clear_default_tags(self) -> None:
        self.provider.default_tags.clear()
        self.default_tags.clear()

    # We now allow customers to bring their own instance
    # of the DatadogProvider provider
    # So we need to define getter/setter for namespace property
    # To access this attribute on the provider instance.
    @property
    def namespace(self):
        return self.provider.namespace

    @namespace.setter
    def namespace(self, namespace):
        self.provider.namespace = namespace



from __future__ import annotations

import dataclasses as dc
import inspect
import pickle
import random
import secrets
import shutil
import sys
import typing as t
import uuid
from pathlib import Path

from boltons.cacheutils import LRU
from pydantic import ValidationError

import datapane._vendor.bottle as bt
from datapane.blocks import Controls, Swap
from datapane.client import log
from datapane.processors import AppTransformations, ConvertXML, Pipeline, PreProcessView, ViewState
from datapane.processors.file_store import GzipTmpFileEntry
from datapane.view import Blocks

if t.TYPE_CHECKING:
    from datapane.common.dp_types import SDict


@dc.dataclass
class FunctionRef:
    f: t.Callable
    controls: Controls
    f_id: str
    swap: Swap = Swap.REPLACE
    # other (decorator) params, ...
    cache: bool = False
    state: bool = dc.field(init=False)
    needs_param: bool = dc.field(init=False)
    # reset: bool

    def __post_init__(self):
        # validate the function
        f_sig = inspect.signature(self.f)
        self.state = "session" in f_sig.parameters

        if "params" in f_sig.parameters:
            self.needs_param = True
        else:
            # check the function has explicit args for all controls
            missing_params = self.controls.param_names - set(f_sig.parameters.keys())
            self.needs_param = bool(missing_params)
            if self.needs_param:
                raise ValueError(
                    f"{self.f} needs either a `params` argument or to handle additional parameters - {missing_params}"
                )

        if self.state or not self.controls.is_cacheable:
            if self.cache:
                log.debug(f"Disabling cache hint for {self.f}")
            self.cache = False

        if self.cache:
            log.debug(f"Enabling experimental cache support for {self.f}")

        log.debug(f"Registered {self.f} ({self.cache=}, {self.state=})")


FEntries = t.Dict[str, FunctionRef]


@dc.dataclass
class GlobalState:
    """Stored within the lifetime of an App and passed through to handlers"""

    app_dir: Path
    main: FunctionRef
    # todo - make threadsafe
    state: SDict = dc.field(default_factory=dict)
    server_id: str = dc.field(default_factory=lambda: uuid.uuid4().hex)
    function_cache: LRU[t.Tuple[str, bytes], ViewState] = dc.field(default_factory=lambda: LRU(max_size=128))

    def clear(self) -> None:
        """Wipe global state - called on app.close"""
        self.function_cache.clear()
        self.state.clear()
        shutil.rmtree(self.app_dir, ignore_errors=True)


@dc.dataclass
class SessionState:
    """Per-user / session state"""

    s_id: str = dc.field(init=False)
    user: str = "anonymous"
    # dict of assets - TODO - type this
    assets: t.Dict = dc.field(default_factory=dict)
    # app function entry points
    entries: FEntries = dc.field(default_factory=dict)
    # internal session storage
    user_state: t.Dict[str, t.Any] = dc.field(default_factory=dict)

    def __post_init__(self):
        self.s_id = secrets.token_urlsafe(18)

    def add_entry(self, app_entry: FunctionRef) -> None:
        """Add the entry to the known entries for the state"""
        self.entries[app_entry.f_id] = app_entry

    def apply_view_state(self, view_state: ViewState) -> None:
        # mutate session state based on latest view update
        self.assets.update(view_state.store.as_dict())
        self.entries.update(view_state.entries)

    def __hash__(self) -> int:
        # TODO - implement a proper hash function for session state
        return random.randint(0, sys.maxsize)


def get_session_state() -> SessionState:
    """Relies on the TLS request object"""
    return t.cast(SessionState, bt.request.session)


def get_global_state() -> GlobalState:
    return ...


def apply_ref(g_s: GlobalState, s_s: SessionState, ref: FunctionRef, params: t.Dict) -> t.Dict:
    """
    Run the referenced app function with the given parameters and return the result
    suitable for encoding in an RPC response
    """
    from .json_rpc import RPCException

    def check_cache() -> t.Optional[t.Dict]:
        # do we have this in the cache??
        if ref.cache:
            _key = (ref.f_id, pickle.dumps(params))
            f_cache = g_s.function_cache
            _val: t.Optional[ViewState] = f_cache.get(_key)
            _hit_ratio = (f_cache.hit_count / (f_cache.soft_miss_count + f_cache.hit_count)) * 100
            log.debug(f"Cache info - {f_cache.soft_miss_count} misses, {f_cache.hit_count} hits ({_hit_ratio:.2f}%)")

            if _val:
                log.debug("Got cache hit for request")
                return build_rpc_res(_val)
        return None

    def store_cache(vs: ViewState) -> None:
        if ref.cache:
            _key = (ref.f_id, pickle.dumps(params))
            g_s.function_cache[_key] = vs

    def gen_args():
        # "massage" the parameters to the function
        try:
            ParamsModel = ref.controls._to_pydantic()
            params_model = ParamsModel(**params)
            in_params: dict = params_model.dict()

        except ValidationError as e:
            # TODO - fix json-rpc id handling here
            raise RPCException(-32602, "Invalid Parameters", data=e.errors()) from e

        # function arg handling / binding
        f_sig = inspect.signature(ref.f, follow_wrapped=True)

        # unpack params
        kwargs = {}
        for k in list(in_params.keys()):
            if k in f_sig.parameters:
                kwargs[k] = in_params.pop(k)

        # add params if needed
        if ref.needs_param:
            kwargs["params"] = in_params
        elif len(in_params) > 0:
            log.warning(f"'params' not present in {ref.f} but additional parameters present - {in_params}, dropping")

        # add optional state param
        if ref.state:  # "state" in f_sig.parameters:
            kwargs["session"] = s_s.user_state

        f_args = f_sig.bind(**kwargs)
        f_args.apply_defaults()
        return f_args

    is_fragment = ref.f_id != "app.main"

    def build_view(res) -> ViewState:
        # Build a view for serving, using the view_ast as a base state and storing assets in dest
        # TODO - use ref.swap to validate return type
        blocks = Blocks.wrap_blocks(res)
        assets_dir = g_s.app_dir / "assets"

        # write the app html and assets
        vs = ViewState(blocks=blocks, file_entry_klass=GzipTmpFileEntry, dir_path=assets_dir)

        return (
            Pipeline(vs)
            .pipe(PreProcessView(is_finalised=False))
            .pipe(AppTransformations())
            .pipe(ConvertXML(fragment=is_fragment))
            .state
        )

    def build_rpc_res(vs: ViewState) -> dict:
        # we need to reapply the viewstate for the specific user session
        s_s.apply_view_state(vs)
        return dict(is_fragment=is_fragment, view_xml=vs.view_xml, assets=vs.store.as_dict())

    ############################################################################
    # Core logic
    if rpc_res := check_cache():
        return rpc_res

    # call the function with the processed input parameters
    f_args = gen_args()
    res = ref.f(*f_args.args, **f_args.kwargs)
    # build the view response
    vs = build_view(res)
    # construct the RPC return
    rpc_res = build_rpc_res(vs)
    store_cache(vs)

    return rpc_res



# flake8: noqa:F401
# import typing as t

from .asset import Attachment, DataTable, Media, Plot, Table
from .base import BaseBlock, BlockList, BlockOrPrimitive, DataBlock, wrap_block
from .compute import Compute, Controls, Dynamic, Empty, Form, Swap, TargetMode, Trigger
from .layout import Group, Page, Select, SelectType, Toggle, VAlign
from .misc_blocks import BigNumber
from .parameters import Choice, Date, DateTime, File, MultiChoice, NumberBox, Range, Switch, Tags, TextBox, Time
from .text import HTML, Code, Embed, Formula, Text

# Block = t.Union["Group", "Select", "DataBlock", "Empty", "Function"]
Block = BaseBlock



import logging
import typing as t
from datetime import timedelta
from enum import Enum
from os import PathLike
from pathlib import Path

# Typedefs
# A JSON-serialisable config object
SDict = t.Dict[str, t.Any]
SSDict = t.Dict[str, str]
SList = t.List[str]
# NOTE - mypy cannot handle recursive types like this currently. Will review in the future
JSON = t.Union[str, int, float, bool, None, t.Mapping[str, "JSON"], t.List["JSON"]]  # type: ignore
JDict = SDict  # should be JSON
JList = t.List[JSON]  # type: ignore
MIME = t.NewType("MIME", str)
URL = t.NewType("URL", str)
HTML = t.NewType("HTML", str)
NPath = t.Union[Path, PathLike, str]
Hash = t.NewType("Hash", str)
EnumType = int  # alias for enum values

# Constants
# NOTE - PKL_MIMETYPE and ARROW_MIMETYPE are custom mimetypes
PKL_MIMETYPE = MIME("application/vnd.pickle+binary")
ARROW_MIMETYPE = MIME("application/vnd.apache.arrow+binary")
ARROW_EXT = ".arrow"
TD_1_HOUR = timedelta(hours=1)
TD_1_DAY = timedelta(days=1)
SECS_1_HOUR: int = int(TD_1_HOUR.total_seconds())
SECS_1_WEEK: int = int(timedelta(weeks=1).total_seconds())
SIZE_1_MB: int = 1024 * 1024


class StrEnum(str, Enum):
    # TODO - replace with StrEnum in py3.11 stdlib
    def __str__(self):
        return str(self.value)


# Errors
class DPError(Exception):
    """Base DP Error"""

    pass


log = logging.getLogger("datapane")



from __future__ import annotations

import base64
from typing import Any, Callable, Dict, List

import botocore
from aws_encryption_sdk import (
    CachingCryptoMaterialsManager,
    EncryptionSDKClient,
    LocalCryptoMaterialsCache,
    StrictAwsKmsMasterKeyProvider,
)

from aws_lambda_powertools.shared.user_agent import register_feature_to_botocore_session
from aws_lambda_powertools.utilities._data_masking.constants import (
    CACHE_CAPACITY,
    MAX_CACHE_AGE_SECONDS,
    MAX_MESSAGES_ENCRYPTED,
)
from aws_lambda_powertools.utilities._data_masking.provider import BaseProvider


class ContextMismatchError(Exception):
    def __init__(self, key):
        super().__init__(f"Encryption Context does not match expected value for key: {key}")
        self.key = key


class AwsEncryptionSdkProvider(BaseProvider):
    """
    The AwsEncryptionSdkProvider is used as a provider for the DataMasking class.

    This provider allows you to perform data masking using the AWS Encryption SDK
    for encryption and decryption. It integrates with the DataMasking class to
    securely encrypt and decrypt sensitive data.

    Usage Example:
    ```
    from aws_lambda_powertools.utilities.data_masking import DataMasking
    from aws_lambda_powertools.utilities.data_masking.providers.kms.aws_encryption_sdk import (
        AwsEncryptionSdkProvider,
    )


    def lambda_handler(event, context):
        provider = AwsEncryptionSdkProvider(["arn:aws:kms:us-east-1:0123456789012:key/key-id"])
        masker = DataMasking(provider=provider)

        data = {
            "project": "powertools",
            "sensitive": "xxxxxxxxxx"
        }

        masked = masker.encrypt(data,fields=["sensitive"])

        return masked

    ```
    """

    def __init__(
        self,
        keys: List[str],
        key_provider=None,
        local_cache_capacity: int = CACHE_CAPACITY,
        max_cache_age_seconds: float = MAX_CACHE_AGE_SECONDS,
        max_messages_encrypted: int = MAX_MESSAGES_ENCRYPTED,
        json_serializer: Callable | None = None,
        json_deserializer: Callable | None = None,
    ):
        super().__init__(json_serializer=json_serializer, json_deserializer=json_deserializer)

        self._key_provider = key_provider or KMSKeyProvider(
            keys=keys,
            local_cache_capacity=local_cache_capacity,
            max_cache_age_seconds=max_cache_age_seconds,
            max_messages_encrypted=max_messages_encrypted,
            json_serializer=self.json_serializer,
            json_deserializer=self.json_deserializer,
        )

    def encrypt(self, data: bytes | str | Dict | int, **provider_options) -> str:
        return self._key_provider.encrypt(data=data, **provider_options)

    def decrypt(self, data: str, **provider_options) -> Any:
        return self._key_provider.decrypt(data=data, **provider_options)


class KMSKeyProvider:

    """
    The KMSKeyProvider is responsible for assembling an AWS Key Management Service (KMS)
    client, a caching mechanism, and a keyring for secure key management and data encryption.
    """

    def __init__(
        self,
        keys: List[str],
        json_serializer: Callable,
        json_deserializer: Callable,
        local_cache_capacity: int = CACHE_CAPACITY,
        max_cache_age_seconds: float = MAX_CACHE_AGE_SECONDS,
        max_messages_encrypted: int = MAX_MESSAGES_ENCRYPTED,
    ):
        session = botocore.session.Session()
        register_feature_to_botocore_session(session, "data-masking")

        self.json_serializer = json_serializer
        self.json_deserializer = json_deserializer
        self.client = EncryptionSDKClient()
        self.keys = keys
        self.cache = LocalCryptoMaterialsCache(local_cache_capacity)
        self.key_provider = StrictAwsKmsMasterKeyProvider(key_ids=self.keys, botocore_session=session)
        self.cache_cmm = CachingCryptoMaterialsManager(
            master_key_provider=self.key_provider,
            cache=self.cache,
            max_age=max_cache_age_seconds,
            max_messages_encrypted=max_messages_encrypted,
        )

    def encrypt(self, data: bytes | str | Dict | float, **provider_options) -> str:
        """
        Encrypt data using the AwsEncryptionSdkProvider.

        Parameters
        -------
            data : Union[bytes, str]
                The data to be encrypted.
            provider_options
                Additional options for the aws_encryption_sdk.EncryptionSDKClient

        Returns
        -------
            ciphertext : str
                The encrypted data, as a base64-encoded string.
        """
        data_encoded = self.json_serializer(data)
        ciphertext, _ = self.client.encrypt(
            source=data_encoded,
            materials_manager=self.cache_cmm,
            **provider_options,
        )
        ciphertext = base64.b64encode(ciphertext).decode()
        return ciphertext

    def decrypt(self, data: str, **provider_options) -> Any:
        """
        Decrypt data using AwsEncryptionSdkProvider.

        Parameters
        -------
            data : Union[bytes, str]
                The encrypted data, as a base64-encoded string
            provider_options
                Additional options for the aws_encryption_sdk.EncryptionSDKClient

        Returns
        -------
            ciphertext : bytes
                The decrypted data in bytes
        """
        ciphertext_decoded = base64.b64decode(data)

        expected_context = provider_options.pop("encryption_context", {})

        ciphertext, decryptor_header = self.client.decrypt(
            source=ciphertext_decoded,
            key_provider=self.key_provider,
            **provider_options,
        )

        for key, value in expected_context.items():
            if decryptor_header.encryption_context.get(key) != value:
                raise ContextMismatchError(key)

        ciphertext = self.json_deserializer(ciphertext)
        return ciphertext



"""
Basic app server implementation, using,
- chroot as the web server
- bottle as a WSGI micro-framework
- custom json-rpc dispatcher
"""
from __future__ import annotations

import atexit
import os
import sys
import tempfile
import threading
import typing as t
from contextlib import suppress
from getpass import getpass
from pathlib import Path

import cheroot.server
import cheroot.ssl.builtin
import cheroot.wsgi
import importlib_resources as ir

import datapane._vendor.bottle as bt
from datapane.blocks import Controls
from datapane.client import log
from datapane.client.analytics import capture
from datapane.common.dp_types import SECS_1_HOUR, SIZE_1_MB
from datapane.common.utils import guess_type
from datapane.ipython.environment import get_environment
from datapane.processors.processors import ExportBaseHTMLOnly
from datapane.view import Blocks, BlocksT

try:
    import ipywidgets
except ImportError:
    ipywidgets = None

from .json_rpc import app_rpc_call
from .plugins import DPBottlePlugin
from .runtime import FunctionRef, GlobalState, get_session_state

# globally set the max request size to 100 MB
bt.BaseRequest.MEMFILE_MAX = int(os.getenv("MAX_REQUEST_BODY", 100 * SIZE_1_MB))


def create_bottle_app(debug: bool) -> bt.Bottle:
    # Create the app and all nested routes
    app: bt.Bottle = bt.Bottle()
    # Update app config
    app.config.update(
        {
            # 'autojson': True,
            # "dp_bottle.session_timeout": SECS_1_HOUR,
            "dp_bottle.CORS": "*",
        }
    )

    # Test API endpoint
    @app.get("/hello/")
    def hello():
        return "Hello World!"

    @app.get("/")
    def initial_app() -> str:
        # TODO - cache this
        # this is kinda hacky, reusing the template system
        # for exporting reports here for the chrome only
        html = ExportBaseHTMLOnly(debug).generate_chrome()
        return html

    bt.debug(debug)
    if debug:
        # enable serving the web-components assets directly from py app server
        root_dir: Path = t.cast(Path, ir.files("datapane"))
        web_dist_dir = root_dir.parent.parent.parent / "web-components" / "dist"

        @app.get("/web-static/<fpath:path>")
        def serve_web_static(fpath):
            # headers = {
            #     "Cache-Control": f"private, max-age={SECS_1_HOUR}, must-revalidate, no-transform, immutable",
            # }
            headers = {}
            return bt.static_file(filename=fpath, root=web_dist_dir, headers=headers)

    return app


# @app.get("/assets/<filename>")
def serve_file(global_state: GlobalState, filename: str):
    _ = get_session_state()

    # pull from the asset store, or directly from the file?
    headers = {
        "Content-Encoding": "gzip",
        "Cache-Control": f"private, max-age={SECS_1_HOUR}, must-revalidate, no-transform, immutable",
    }
    root = global_state.app_dir / "assets"
    mime = guess_type(root / filename)
    log.info(f"Serving asset file {filename=} ({mime=})")
    return bt.static_file(filename=filename, root=root, mimetype=mime, headers=headers)


# See tests/client/test_server.py for important info on testing this code.
def _build_app(entry: Blocks, *, debug: bool = False, embed_mode: bool = False) -> bt.Bottle:
    # setup GlobalState
    app_dir = Path(tempfile.mkdtemp(prefix="dp-"))
    assets_dir = app_dir / "assets"
    assets_dir.mkdir(parents=True)

    # hardcode the main app entry
    entry_f = lambda params: entry  # noqa: E731
    cache = True
    main_entry = FunctionRef(f=entry_f, controls=Controls.empty(), f_id="app.main", cache=cache)
    g_s = GlobalState(app_dir, main_entry)

    # create app and add routes
    app = create_bottle_app(debug)
    app.route("/app-rpc-call/", ["POST"], lambda: app_rpc_call(g_s))
    app.route("/assets/<filename>", ["GET"], lambda filename: serve_file(g_s, filename))
    app.route("/control/reset/", ["POST"], lambda: None)

    # start the server
    app.config.update({})
    app.install(DPBottlePlugin(g_s, embed_mode))

    if debug is not None:
        bt._debug(debug)

    return app


def serve_app(
    entry: BlocksT,
    *,
    port: t.Optional[int] = None,
    host: t.Optional[str] = None,
    debug: bool = False,
    ui: t.Optional[str] = None,
    public: bool = False,
    embed_mode: bool = False,
) -> None:
    """
    Main app serve entrypoint.

    Args:
        port: Select a specific port, otherwise one will be chosen automatically.
        host: Specify a specific host.
        debug: Enable debug mode.
        ui: Select a specific UI to use (`console` or `ipywidgets`).
        public: Expose the app to the Internet with ngrok (requires auth_token).
        embed_mode: Allows embedding in some third-party products by disabling security policies which will not work in an iframe.
    """
    wrapped_entry = Blocks.wrap_blocks(entry)
    if not wrapped_entry.has_compute:
        log.warning("Running serve_app on a static report - do you want to use save_report or upload_report instead?")

    app = _build_app(wrapped_entry, debug=debug, embed_mode=embed_mode)
    capture("App Server Started")

    if port is None:
        if _port := os.environ.get("PORT"):
            port = int(_port)

            # when the port is provided from the environment, it likely means
            # we're running in a hosted environment.
            # Bind to all interfaces so the proxy can route to us.
            if host is None:
                host = "0.0.0.0"
    if host is None:
        host = "127.0.0.1"

    server = CherootServer(app, host=host, port=port, preferred_port=8080, quiet=False)
    ui_cls: t.Type[ControllerUI]
    if ui is None:
        ui_cls = _choose_ui_class()
    else:
        if ui not in CONTROLLER_UIS:
            raise ValueError(f"UI choice '{ui}' is not a known option. Choose from: {', '.join(CONTROLLER_UIS.keys())}")
        ui_cls = CONTROLLER_UIS[ui]
    controller = ServerController(server, cleanups=[app.close], ui_cls=ui_cls, public=public)
    controller.go()


MAX_ACQUIRE_PORT_ATTEMPTS = 1000


def _choose_ui_class() -> t.Type[ControllerUI]:
    env = get_environment()
    if env.is_notebook_environment:
        if env.supports_ipywidgets:
            if ipywidgets is not None:
                return IPyWidgetsControllerUI
            else:
                print(
                    "Please install Jupyter Widgets for an improved experienced with apps in Notebook environments: "
                    + "https://ipywidgets.readthedocs.io/en/latest/user_install.html",
                    file=sys.stderr,
                )
                return FallbackNotebookControllerUI
        else:
            print(
                "You can run this notebook in Jupyter Lab or Jupyter Notebook for an improved experience.\n",
                file=sys.stderr,
            )
            return FallbackNotebookControllerUI

    else:
        return ConsoleControllerUI


# Server/controller/UI classes:

# CherootServer
# -------------
#
# This class is similar to bottle's ServerAdapter abstraction, but with our own
# needs, specifically:
#
# - automatic port numbering
# - communicates with a controller to display messages

# Internally it wraps the `cheroot.wsgi.Server` to provide all HTTP functionality


# ServerController
# ----------------

# Responsible for starting and stopping the server. Connects with both
# CherootServer and the UI class

# ControllerUI
# ------------
#
# Abstract, has implementations for console, ipywidgets and other notebook fallbacks
#
# Responsible for displaying messages to the user, and "Stop server" controls.
# Communicates with ServerController.


class CherootHTTPRequest(cheroot.server.HTTPRequest):
    def ensure_headers_sent(self, *args, **kwargs):
        try:
            super().ensure_headers_sent(*args, **kwargs)
        except OSError:
            # We're at a low enough level to know that OSError means the socket
            # closed, likely from a timeout.
            # bail out silently, to let Cheroot handle response cleanup correctly.
            # This also keeps the messy traceback from showing.
            pass


class CherootHTTPConnection(cheroot.server.HTTPConnection):
    RequestHandlerClass = CherootHTTPRequest


class CherootServer:
    quiet = False

    def __init__(
        self,
        handler: t.Callable,
        host: str = "127.0.0.1",
        port: t.Optional[int] = None,
        preferred_port: t.Optional[int] = None,
        quiet: bool = False,
        **options,
    ):
        """
        Pass port to specify an exact port, or None for automatic
        Pass preferred_port to specify a port that you prefer, but accept others.
        """

        self.handler = handler
        self.options = options
        # force IPv4 for localhost
        # self.host = "127.0.0.1" if host.strip().lower() == "localhost" else host
        self.host = host.strip().lower()
        self.port = port
        self.preferred_port = preferred_port
        self.quiet = quiet
        self.server: cheroot.wsgi.Server | None = None

    def _make_server(self, port: int) -> None:
        options = self.options.copy()
        options["bind_addr"] = (self.host, port)
        options["wsgi_app"] = self.handler
        certfile = options.pop("certfile", None)
        keyfile = options.pop("keyfile", None)
        chainfile = options.pop("chainfile", None)
        server = cheroot.wsgi.Server(**options)
        server.ConnectionClass = CherootHTTPConnection
        if certfile and keyfile:
            server.ssl_adapter = cheroot.ssl.builtin.BuiltinSSLAdapter(certfile, keyfile, chainfile)
        server.prepare()  # This will raise an error on failure to bind port
        self.server = server
        # Having successfully acquired a port, copy it back to self.
        self.host = server.bind_addr[0]
        self.port = server.bind_addr[1]

    @property
    def url(self) -> str:
        if self.host.startswith("unix:"):
            return self.host
        elif ":" in self.host:  # IPv6
            return f"http://[{self.host}]:{self.port}"
        else:
            return f"http://{self.host}:{self.port}"

    def acquire_port(self):
        """
        Bind to the requested or an automatic port. Raises an exception if not possible.
        """
        if self.port is None:
            if self.preferred_port is None:
                # Fully automatic, any port will do.
                desired_port = 0
                self._make_server(desired_port)
            else:
                # Automatic, starting from preferred_port
                desired_port = self.preferred_port
                # Start with desired_port, or next available if it is taken
                for i in range(0, MAX_ACQUIRE_PORT_ATTEMPTS):
                    try:
                        self._make_server(desired_port)
                    except OSError:
                        desired_port += 1
                    else:
                        break
                if self.server is None:
                    # None of the attempts succeeded
                    raise
        else:
            self._make_server(self.port)

    def serve(self):  # pragma: no cover
        try:
            self.server.serve()
        finally:
            self.server.stop()

    def stop(self):
        self.server.stop()


class ControllerUI:
    server_needs_background_thread: bool = False

    def __init__(self, controller: ServerController) -> None:
        self.controller = controller

    def display(self):
        """
        Display the UI
        """
        raise NotImplementedError()

    def show_message(self, value: str):
        """
        Print a message onto the UI
        """
        raise NotImplementedError()


class ServerController:
    def __init__(
        self, server: CherootServer, *, cleanups: list[t.Callable] = None, ui_cls: t.Type[ControllerUI], public: bool
    ) -> None:
        self.server: CherootServer = server
        self.cleanups: list[t.Callable] = cleanups or []
        self.thread: threading.Thread = None
        self.ui: ControllerUI = ui_cls(self)
        self.public = public

    def go(self):
        self.server.acquire_port()
        if self.public:
            self.start_ngrok()

        self.ui.display()

        if self.ui.server_needs_background_thread:
            self.run_server_in_thread()
        else:
            self.run_server()

    def start_ngrok(self):
        from pyngrok import ngrok

        current_config = {}
        config_path = ngrok.conf.get_default().config_path or ngrok.conf.DEFAULT_NGROK_CONFIG_PATH

        if config_path:
            with suppress(FileNotFoundError):
                current_config = ngrok.installer.get_ngrok_config(config_path, use_cache=False)

        if current_config.get("authtoken", None):
            ngrok_auth_token = current_config.get("authtoken")
        else:
            ngrok_auth_token = getpass("Enter ngrok auth token (see https://ngrok.com): ")

        ngrok.set_auth_token(ngrok_auth_token)
        http_tunnel = ngrok.connect(self.server.port, bind_tls=True)
        self.ui.show_message(f"Public URL: {http_tunnel.public_url}\n")
        self.cleanups.append(ngrok.kill)

    def run_server(self):
        try:
            self.server.serve()
        except KeyboardInterrupt:
            pass
        finally:
            self.do_cleanups()

    def run_server_in_thread(self):
        assert self.thread is None
        thread = threading.Thread(target=self.server.serve, daemon=True)
        thread.start()
        self.thread = thread
        # To cover the case of non-clean shutdown:
        atexit.register(self.do_cleanups)

    def stop_background_server(self):
        # This is only used for threaded mode
        self.server.stop()
        if self.thread is not None:
            self.thread.join()
        self.do_cleanups()
        atexit.unregister(self.do_cleanups)

    def do_cleanups(self):
        for cleanup in self.cleanups:
            cleanup()

    @property
    def server_url(self) -> str:
        return self.server.url


class ConsoleControllerUI(ControllerUI):
    server_needs_background_thread = False

    def display(self):
        self.show_message(f"App running on {self.controller.server_url}/\n")
        self.show_message("Hit Ctrl-C to quit.\n")

    def show_message(self, value: str):
        print(value, file=sys.stderr, end="")


class FallbackNotebookControllerUI(ControllerUI):
    server_needs_background_thread = False

    def display(self):
        from IPython.display import HTML, display

        server_url = self.controller.server_url
        display(
            HTML(
                f'<p>Your app is running on: <a href="{server_url}">{server_url}</a></p>'
                + "<p>Use notebook ⏹ or 'Stop/Interrupt' button to halt.</p>"
            )
        )

    def show_message(self, value: str):
        # Autolink URLs by treating as Markdown
        from IPython.display import Markdown, display

        display(Markdown(value))


if ipywidgets is not None:

    class WarningWidget(ipywidgets.VBox):
        """
        Container widget that shows a fallback message for the case where Jupyter Widgets is not installed.
        """

        def __init__(self, child, html_message):
            super().__init__(children=[child])
            self.html_message = html_message

        def _repr_html_(self):
            return self.html_message


class IPyWidgetsControllerUI(ControllerUI):
    server_needs_background_thread = True

    def __init__(self, controller: ServerController) -> None:
        super().__init__(controller)

        self.button = ipywidgets.Button(description="Stop app server", icon="stop")
        self.output = ipywidgets.Output()

        def on_button_clicked(b):
            self.show_message("Stopping server...")
            self.controller.stop_background_server()
            self.show_message("stopped.\n")
            self.show_message("Re-run cell to run server again.\n")
            self.button.disabled = True
            self.button.description = "Stopped"

        self.button.on_click(on_button_clicked)

    def show_message(self, value: str):
        self.output.append_stdout(value)

    def display(self):
        from IPython.display import display

        server_url = self.controller.server_url
        display(
            WarningWidget(
                ipywidgets.VBox([self.button, self.output]),
                (
                    "<ul>"
                    + f'  <li>Your app is running on: <a href="{server_url}">{server_url}</a></li>'
                    + "   <li>You're seeing this message because 'Jupyer Widgets' is not installed in your Jupyter notebook/lab environment. "
                    + "    so the UI for controlling the app server is not visible. See "
                    + '    <a href="https://ipywidgets.readthedocs.io/en/latest/user_install.html">Jupyter Widgets Installation instructions</a></li>'
                    + '  <li>Use <code>dp.serve_app(ui="console")</code> to use the console UI instead.</li>'
                    + "</ul>"
                ),
            )
        )
        self.show_message(f"App running on {server_url}/\n")


CONTROLLER_UIS = {
    "console": ConsoleControllerUI,
    "ipywidgets": IPyWidgetsControllerUI,
}



import types
from enum import Enum
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional, Set, Type, Union

if TYPE_CHECKING:
    from pydantic import BaseModel  # noqa: F401

CacheKey = Optional[Callable[..., Any]]
IncEx = Union[Set[int], Set[str], Dict[int, Any], Dict[str, Any]]
ModelNameMap = Dict[Union[Type["BaseModel"], Type[Enum]], str]
TypeModelOrEnum = Union[Type["BaseModel"], Type[Enum]]
UnionType = getattr(types, "UnionType", Union)


COMPONENT_REF_PREFIX = "#/components/schemas/"
COMPONENT_REF_TEMPLATE = "#/components/schemas/{model}"
METHODS_WITH_BODY = {"GET", "HEAD", "POST", "PUT", "DELETE", "PATCH"}


validation_error_definition = {
    "title": "ValidationError",
    "type": "object",
    "properties": {
        "loc": {
            "title": "Location",
            "type": "array",
            "items": {"anyOf": [{"type": "string"}, {"type": "integer"}]},
        },
        "msg": {"title": "Message", "type": "string"},
        "type": {"title": "Error Type", "type": "string"},
    },
    "required": ["loc", "msg", "type"],
}

validation_error_response_definition = {
    "title": "HTTPValidationError",
    "type": "object",
    "properties": {
        "detail": {
            "title": "Detail",
            "type": "array",
            "items": {"$ref": COMPONENT_REF_PREFIX + "ValidationError"},
        },
    },
}



DATA_TYPES = {
    "Mutation": [
        {
            "format": ["maf"],
            "supported_repo": [
                {
                    "name": "cbioportal",
                    "header_mapping": {
                        "gene": "Hugo_Symbol",
                        "chr": "Chromosome",
                        "startPosition": "Start_Position",
                        "endPosition": "End_Position",
                        "referenceAllele": "Reference_Allele",
                        "variantAllele": "Tumor_Seq_Allele2",
                        "mutationType": "Variant_Classification",
                        "variantType": "Variant_Type",
                        "uniqueSampleKey": "Tumor_Sample_Barcode",
                    },
                },
                {"name": "tcga", "header_mapping": {}},
            ],
        }
    ]
}
COHORT_SUPPORTED_DATATYPES = ["Raw Counts Transcriptomics"]
COHORT_SUPPORTED_DATASOURCES = ["GEO"]
COHORT_LIST_COLS_TO_DROP = [
    "sample_characteristics",
    "timestamp_",
    "curated_control",
    "curated_cohort_id",
    "curated_cohort_name",
    "curated_is_control",
    "curated_max_age",
    "curated_min_age",
    "curated_age_unit",
]
MA_SUPPORTED_REPO = ["geo_transcriptomics_omixatlas"]

# validation flow files url github
VALIDATION_FLOW_FILES_URL = (
    "https://raw.githubusercontent.com/ElucidataInc/PublicAssets/master/"
    + "internal-user/validation_full_flow_files"
)

# POLLY_PY_TEST_FILES_URL = (
#     "https://raw.githubusercontent.com/ElucidataInc/PublicAssets/master/"
#     + "internal-user/polly_py_test_files"
# )

# # repo id used in the tests
# INGESTION_TEST_1_REPO_ID = "1654268055800"

# endpoints
CONSTANTS_ENDPOINT = "/constants"
REPOSITORIES_ENDPOINT = "/repositories"
REPOSITORY_PACKAGE_ENDPOINT = REPOSITORIES_ENDPOINT + "/{}/packages"
IMAGE_URL_ENDPOINT = (
    "https://elucidatainc.github.io/PublicAssets/discover-fe-assets/omixatlas_hex.svg"
)

# statuscodes
OK = 200
CREATED = 201
COMPUTE_ENV_VARIABLE = "POLLY_TYPE"
UPLOAD_URL_CREATED = 204
UPLOAD_ERROR_CODE = 400

# Multipart S3 upload parameters
KB = 1024
MB = KB * KB
GB = 1024 * MB
SMALL_FILE_SIZE = 20 * GB
MEDIUM_FILE_SIZE = 40 * GB

MULTIPART_THRESHOLD = 25 * MB
MAX_CONCURRENCY = 32

# Chunksize for file less than 20 gb
MULTIPART_CHUNKSIZE_SMALL_FILE_SIZE = 25 * MB
IO_CHUNKSIZE_SMALL_FILE_SIZE = 25 * MB

# Chunksize for file more than 20gb and less than 40gb
MULTIPART_CHUNKSIZE_MEDIUM_FILE_SIZE = 50 * MB
IO_CHUNKSIZE_MEDIUM_FILE_SIZE = 50 * MB

# Chunksize for file more than 40gb
MULTIPART_CHUNKSIZE_LARGE_FILE_SIZE = 100 * MB
IO_CHUNKSIZE_LARGE_FILE_SIZE = 100 * MB

# S3 Exceptions
EXPIRED_TOKEN = "ExpiredToken"

# cohort constants
COHORT_VERSION = "0.2"
COHORT_CONSTANTS_URL = (
    "https://elucidatainc.github.io/PublicAssets/cohort_constants.txt"
)

OBSOLETE_METADATA_FIELDS = [
    "package",
    "region",
    "bucket",
    "key",
    "file_type",
    "file_location",
    "src_uri",
    "timestamp_",
]
dot = "."

GETTING_UPLOAD_URLS_PAYLOAD = {"data": {"type": "files", "attributes": {"folder": ""}}}

INGESTION_LEVEL_METADATA = {
    "id": "metadata/ingestion",
    "type": "ingestion_metadata",
    "attributes": {
        "ignore": "false",
        "urgent": "true",
        "v1_infra": False,
        "priority": "low",
    },
}

METADATA = {"data": []}

GET_SCHEMA_RETURN_TYPE_VALS = ["dataframe", "dict"]

COMBINED_METADATA_FILE_NAME = "combined_metadata.json"

FILES_PATH_FORMAT = {"metadata": "<metadata_path>", "data": "<data_path>"}

INGESTION_FILES_PATH_DIR_NAMES = ["metadata", "data"]

SCHEMA_CONFIG_KEY_NAMES = ["source", "datatype"]

VALIDATION_SCOPE_CONSTANTS = ["basic", "advanced"]

SCHEMA_CONFIG_FORMAT = {"source": "<source_name", "datatype": "<datatype_name>"}

FORMATTED_METADATA = {"id": "", "type": "", "attributes": {}}

FILE_FORMAT_CONSTANTS_URL = (
    "https://elucidatainc.github.io/PublicAssets/file_format_constants.txt"
)

VALIDATION_STATUS_FILE_NAME = "validation_status.txt"

# this constant is used as a base url for getting data files for
# testing ingestion functionality
BASE_TEST_FORMAT_CONSTANTS_URL = (
    "https://raw.githubusercontent.com/ElucidataInc/PublicAssets/master/"
    + "internal-user/add_dataset_test_file"
)

ENCRYPTION_KEY_URL = (
    "https://raw.githubusercontent.com/ElucidataInc/PublicAssets/"
    + "56440d3be03afe700af65d41e7fdf610befccce9/key/Encryptionkey.key"
)

COMPRESSION_TYPES = [
    ".br," ".bz2",
    ".bz",
    ".gz",
    ".lz",
    ".lz4",
    ".sz",
    ".rz",
    ".xz",
    ".zip",
    ".tar",
    ".bgz",
]

FILE_FORMAT_CONSTANTS = {
    "data": [
        ".gct",
        ".vcf",
        ".h5ad",
        ".mmcif",
        ".h5seurat",
        ".biom",
        ".zip",
        ".fs",
        ".tar.gz",
        ".fcs",
        ".vcf.bgz",
        ".gct.bz",
        ".csv",
    ],
    "metadata": [".json", ".jpco"],
}
NOT_NEEDED_SCHEMA_FIELDS = [
    "is_current",
    "data_table_version",
    "data_table_name",
    "timestamp_",
]


OMIXATLAS_CATEGORY_VALS = ["private", "public", "diy_poa"]
OMIXATLAS_DATA_TYPE_VALS = ["single_cell", "bulk_rna_seq"]

VALIDATION_LEVEL_CONSTANTS = {"advanced": "value", "basic": "schema"}


SCHEMA_WRITE_SUCCESS = (
    "Job for your schema update has been recorded. Depending on the changes,"
    + "minor updates or complete reindexing will be performed for you data. The repository will be locked for the "
    + "duration of this update. Please refer to the ingestion monitoring dashboard for visibility on this process."
)

MOVE_DATA_SUCCESS = (
    "Your ingestion request is successfully logged. "
    + "You can go to ingestion monitoring dashboard for tracking it's status."
)

SCHEMA_INSERT_SUCCESS = "Schema has been Inserted. Please use get_schema functionality to get the inserted schema."

DATA_COMMIT_MESSAGE = (
    "Your request is successfully logged. "
    + "You can go to ingestion monitoring dashboard for tracking it's status."
)

WAIT_FOR_COMMIT = (
    "Please wait for 30 seconds while your ingestion request is getting logged."
)

WAIT_FOR_COMMIT_DELETE = (
    "Please wait for 30 seconds while your deletion request is getting logged."
)


SCHEMA_UPDATE_SUCCESS = (
    "The schema update is in progress. It should only take a few minutes and you can track the status"
    + " of this update in the ingestion monitoring dashboard. The repository will be locked for this time "
    + " and any further changes in the schema or properties of omixatlas can not be done."
)

SCHEMA_REINGEST_SUCCESS = (
    "The data needs to be re-ingested to accommodate the changes made in the schema. This may "
    + "take from a few minutes to a few hours depending on the number of datasets in the atlas. The"
    + " repository will be locked for this time and any further changes in the schema or properties of "
    + " omixatlas can not be done. Please refer to the ingestion monitoring dashboard to check the"
    + " status of re-indexing."
)

SCHEMA_NO_CHANGE_SUCCESS = (
    "Schema has been updated. Use get schema to view the updated schema"
)

SCHEMA_UPDATE_GENERIC_MSG = "Schema update is in progress. Please refer to ingestion monitoring dashboard for more details."

SCHEMA_REPLACE_GENERIC_MSG = "Schema replace is in progress. Please refer to ingestion monitoring dashboard for more details."

SCHEMA_UPDATE = "schema_update"
SCHEMA_REINGESTION = "reingestion"
SCHEMA_NO_CHANGE = "no_change"

SCHEMA_VERDICT_DICT = {
    "schema_update": SCHEMA_UPDATE_SUCCESS,
    "reingestion": SCHEMA_REINGEST_SUCCESS,
    "no_change": SCHEMA_NO_CHANGE_SUCCESS,
}

VALIDATION_NOT_EXECUTED = (
    "Validation not executed on any of the metadata files.  "
    + "Possible reasons to help you troubleshoot :-"
    + "\n"
    + "All the metadata file passed may have `validate flag in the `__index__` dict to be "
    + "false. Plese set them to true for the files which needs to be validated."
)

DELETION_OPERATION_UNSUCCESSFUL_FOR_ALL_DATASET_IDS = (
    "Files not deleted. Possible reasons to help you troubleshoot :- "
    + "\n"
    + "1. All the datasets passed to delete might not be "
    + "present in the Omixatlas. Please check warning message logged during execution. "
    + "\n"
    + "2. Paths passed for dataset id may be wrong for all the "
    + "dataset ids. Please correct them and try again. Use get_all_file_paths function to "
    + "to get all file_paths corresponding to a dataset_id in Omixatlas. "
    + "\n"
    + "3. Dataset_id was not passed in the list of dataset_ids to be deleted. "
    + "\n"
    + "In case of any other queries, please connect with polly.support@elucidata.io"
)

FIELD_NAME_LOC = "field_name"

DDL_CONST_LIST = [
    "ALL",
    "ALTER",
    "AND",
    "ARRAY",
    "AS",
    "AUTHORIZATION",
    "BETWEEN",
    "BIGINT",
    "BINARY",
    "BOOLEAN",
    "BOTH",
    "BY",
    "CASE",
    "CASHE",
    "CAST",
    "CHAR",
    "COLUMN",
    "CONF",
    "CONSTRAINT",
    "COMMIT",
    "CREATE",
    "CROSS",
    "CUBE",
    "CURRENT",
    "CURRENT_DATE",
    "CURRENT_TIMESTAMP",
    "CURSOR",
    "DATABASE",
    "DATE",
    "DAYOFWEEK",
    "DECIMAL",
    "DELETE",
    "DESCRIBE",
    "DISTINCT",
    "DOUBLE",
    "DROP",
    "ELSE",
    "END",
    "EXCHANGE",
    "EXISTS",
    "EXTENDED",
    "EXTERNAL",
    "EXTRACT",
    "FALSE",
    "FETCH",
    "FLOAT",
    "FLOOR",
    "FOLLOWING",
    "FOR",
    "FOREIGN",
    "FROM",
    "FULL",
    "FUNCTION",
    "GRANT",
    "GROUP",
    "GROUPING",
    "HAVING",
    "IF",
    "IMPORT",
    "IN",
    "INNER",
    "INSERT",
    "INT",
    "INTEGER",
    "INTERSECT",
    "INTERVAL",
    "INTO",
    "IS",
    "JOIN",
    "LATERAL",
    "LEFT",
    "LESS",
    "LIKE",
    "LOCAL",
    "MACRO",
    "MAP",
    "MORE",
    "NONE",
    "NOT",
    "NULL",
    "NUMERIC",
    "OF",
    "ON",
    "ONLY",
    "OR",
    "ORDER",
    "OUT",
    "OUTER",
    "OVER",
    "PARTIALSCAN",
    "PARTITION",
    "PERCENT",
    "PRECEDING",
    "PRECISION",
    "PRESERVE",
    "PRIMARY",
    "PROCEDURE",
    "RANGE",
    "READS",
    "REDUCE",
    "REGEXP",
    "REFERENCES",
    "REVOKE",
    "RIGHT",
    "RLIKE",
    "ROLLBACK",
    "ROLLUP",
    "ROW",
    "ROWS",
    "SELECT",
    "SET",
    "SMALLINT",
    "START",
    "TABLE",
    "TABLESAMPLE",
    "THEN",
    "TIME",
    "TIMESTAMP",
    "TO",
    "TRANSFORM",
    "TRIGGER",
    "TRUE",
    "TRUNCATE",
    "UNBOUNDED",
    "UNION",
    "UNIQUEJOIN",
    "UPDATE",
    "USER",
    "USING",
    "UTC_TIMESTAMP",
    "VALUES",
    "VARCHAR",
    "VIEWS",
    "WHEN",
    "WHERE",
    "WINDOW",
    "WITH",
]

DML_CONST_LIST = [
    "ALTER",
    "AND",
    "AS",
    "BETWEEN",
    "BY",
    "CASE",
    "CAST",
    "CONSTRAINT",
    "CREATE",
    "CROSS",
    "CUBE",
    "CURRENT_DATE",
    "CURRENT_PATH",
    "CURRENT_TIME",
    "CURRENT_TIMESTAMP",
    "CURRENT_USER",
    "DEALLOCATE",
    "DELETE",
    "DESCRIBE",
    "DISTINCT",
    "DROP",
    "ELSE",
    "END",
    "ESCAPE",
    "EXCEPT",
    "EXECUTE",
    "EXISTS",
    "EXTRACT",
    "FALSE",
    "FIRST",
    "FOR",
    "FROM",
    "FULL",
    "GROUP",
    "GROUPING",
    "HAVING",
    "IN",
    "INNER",
    "INSERT",
    "INTERSECT",
    "INTO",
    "IS",
    "JOIN",
    "LAST",
    "LEFT",
    "LIKE",
    "LOCALTIME",
    "LOCALTIMESTAMP",
    "NATURAL",
    "NORMALIZE",
    "NOT",
    "NULL",
    "OF",
    "ON",
    "OR",
    "ORDER",
    "OUTER",
    "PREPARE",
    "RECURSIVE",
    "RIGHT",
    "ROLLUP",
    "SELECT",
    "TABLE",
    "THEN",
    "TRUE",
    "UNESCAPE",
    "UNION",
    "UNNEST",
    "USING",
    "VALUES",
    "WHEN",
    "WHERE",
    "WITH",
]

# curation_library constants
SUPPORTED_ENTITY_TYPES = [
    "disease",
    "drug",
    "species",
    "tissue",
    "cell_type",
    "cell_line",
    "gene",
]

CURATION_COHORT_CACHE = "./.cache/"

SCHEMA_VALIDATION_BASE_URL = (
    "https://raw.githubusercontent.com/ElucidataInc/PublicAssets/master/"
    + "internal-user/schema_validation"
)

SCHEMA_VALIDATION = {
    "empty_repo_id": f"{SCHEMA_VALIDATION_BASE_URL}/demo_oa_2_empty_repo_id.json",
    "missing_repo_id": f"{SCHEMA_VALIDATION_BASE_URL}/demo_oa_2_missing_repo_id.json",
    "missing_schema_key": f"{SCHEMA_VALIDATION_BASE_URL}/demo_oa_2_missing_schema_key.json",
    "wrong_repo_id": f"{SCHEMA_VALIDATION_BASE_URL}/demo_oa_2_wrong_repo_id.json",
    "field_name_cap": f"{SCHEMA_VALIDATION_BASE_URL}/demo_oa_2_field_name_capital.json",
    "field_name_underscore": f"{SCHEMA_VALIDATION_BASE_URL}/demo_oa_2_field_name_with_underscore.json",
    "field_name_resv_keyword": f"{SCHEMA_VALIDATION_BASE_URL}/demo_oa_2_field_name_with_resv_keyword.json",
    "original_name_empty": f"{SCHEMA_VALIDATION_BASE_URL}/demo_oa_2_original_name_empty.json",
    "original_name_grtr_50": f"{SCHEMA_VALIDATION_BASE_URL}/demo_oa_2_original_name_gtr_50.json",
    "type_cosco": f"{SCHEMA_VALIDATION_BASE_URL}/demo_oa_2_type_cosco.json",
    "is_arr_str": f"{SCHEMA_VALIDATION_BASE_URL}/demo_oa_2_is_arr_str.json",
    "is_keyword_str": f"{SCHEMA_VALIDATION_BASE_URL}/demo_oa_2_is_keywrd_str.json",
    "filter_size_less": f"{SCHEMA_VALIDATION_BASE_URL}/demo_oa_2_filter_size_less.json",
    "filter_size_greater": f"{SCHEMA_VALIDATION_BASE_URL}/demo_oa_2_filter_size_greater.json",
    "display_name_empty": f"{SCHEMA_VALIDATION_BASE_URL}/demo_oa_2_display_name_empty.json",
    "display_name_grtr_50": f"{SCHEMA_VALIDATION_BASE_URL}/demo_oa_2_display_name_grtr_50.json",
    "is_keywrd_is_filter": f"{SCHEMA_VALIDATION_BASE_URL}/demo_oa_2_is_keywrd_is_filter.json",
    "is_keywrd_is_ontology": f"{SCHEMA_VALIDATION_BASE_URL}/demo_oa_2_field_name_is_filter_is_ontology.json",
    "positive_case": f"{SCHEMA_VALIDATION_BASE_URL}/positive_case.json",
    "original_name_int": f"{SCHEMA_VALIDATION_BASE_URL}/original_name_int.json",
    "filter_size_str": f"{SCHEMA_VALIDATION_BASE_URL}/filter_size_str.json",
    "repo_id_and_id_diff": f"{SCHEMA_VALIDATION_BASE_URL}/demo_oa_2_repo_id_and_id_diff.json",
}

MIN_REQUIRED_KEYS_FOR_JOBS = ["cpu", "memory", "machineType", "image", "tag", "name"]

MACHINES_FOR_JOBS = (
    {
        "gp": "4 vCPU, 16GB RAM",
        "ci2xlarge": "16 vCPU, 32GB RAM",
        "ci3xlarge": "36 vCPU, 72GB RAM",
        "mi2xlarge": "4 vCPU, 32GB RAM",
        "mi3xlarge": "8 vCPU, 64GB RAM",
        "mi4xlarge": "16 vCPU, 122GB RAM",
        "mi5xlarge": "32 vCPU, 250GB RAM",
        "mi6xlarge": "64 vCPU, 500GB RAM",
        "mi7xlarge": "64 vCPU, 970GB RAM",
        "mix5xlarge": "16vCPU, 512GB RAM",
        "mix6xlarge": "24vCPU, 768GB RAM",
        "mix7xlarge": "64vCPU, 1024GB RAM",
        "gpusmall": "1 GPU, 4 vCPU, 16GB RAM",
        "gpumedium": "4 GPU, 32 vCPU, 240GB RAM",
        "gpularge": "8 GPU, 64 vCPU, 480GB RAM",
        "gpuxlarge": "8 GPU, 96 vCPU, 760GB RAM",
    },
)

DATA_LOG_MODES = ["latest", "all"]
MIXPANEL_KEY = "91fa77fcf07f7b672b5c5c6c09d8a14c"

PROD_ENV_NAME = "polly"

POLLY_PYTHON_PYPI_URL = "https://pypi.python.org/pypi/polly-python/json"
ERROR_MSG_GET_METADATA = "Argument 'table_name' not valid, please query 'SHOW TABLES IN repo_name' \
using query_metadata function and pick the appropriate table name you want the metadata \
for Example: samples, features, samples_singlecell etc. \
Currently, get_metadata works only for sample level table of gct and h5ad files. \
If you want to get metadata for other tables, please contact polly.support@elucidata.io"

TABLE_NAME_SAMPLE_LEVEL_INDEX_MAP = {
    "samples": "gct_col_metadata",
    "samples_singlecell": "h5ad_col_metadata",
}
POLLY_PYTHON_LATEST_VERSION_FILE = (
    "https://elucidatainc.github.io/PublicAssets/polly_python_latest_version.txt"
)

QUERY_MAX_RECORD_SIZE = 20000
# this control lever is for testing as data is less on testpolly
DEFAULT_PAGE_SIZE_LIST_FILES = 2600

DEFAULT_PAGE_SIZE_LIST_FILES_KEY_NAME = "default_page_size_list_files"

PAGE_REDUCTION_PERCENTAGE_LIST_FILES = 20

PAGE_REDUCTION_PERCENTAGE_LIST_FILES_KEY_NAME = "page_reduction_percentage_list_files"

LIST_FILE_API_CONTROL_LEVER_LINK = (
    "https://elucidatainc.github.io/PublicAssets/list_files_api_control_lever/"
    + "list_files_control_lever.json"
)
QUERY_API_V2 = "v2"

REPOSITORY_PAYLOAD = {
    "data": {
        "type": "repositories",
        "attributes": {
            "frontend_info": {
                "description": "<DESCRIPTION>",
                "display_name": "<REPO_DISPLAY_NAME>",
                "explorer_enabled": True,
                "initials": "<INITIALS>",
            },
            "indexes": {
                "csv": "<REPO_NAME>_csv",
                "files": "<REPO_NAME>_files",
                "gct_data": "<REPO_NAME>_gct_data",
                "gct_metadata": "<REPO_NAME>_gct_metadata",
                "h5ad_data": "<REPO_NAME>_h5ad_data",
                "h5ad_metadata": "<REPO_NAME>_h5ad_metadata",
                "ipynb": "<REPO_NAME>_ipynb",
                "json": "<REPO_NAME>_json",
            },
            "repo_name": "<REPO_NAME>",
            "category": "<CATEGORY>",
        },
    }
}



import os
import ssl
import json
import base64
from polly import helpers
from tqdm import tqdm
import urllib.request
import logging
from polly.omixatlas import OmixAtlas
from polly.errors import (
    InvalidCohortNameException,
    paramException,
    InvalidParameterException,
    InvalidPathException,
    InvalidDatasetException,
    TechnicalFaultException,
    InvalidDatatypeException,
    IncompatibleDataSource,
    error_handler,
)
from polly.constants import COHORT_VERSION, dot
from joblib import Parallel, delayed
import datetime
from cmapPy.pandasGEXpress.parse import parse
from polly.core_cohort import CohortFileStandard


class CohortRepoClass:
    """
    The Class responsible for Encapsulating Cohort Bridge Functionalities.
    These are the functionalities that extend other useful features from OmixAtlas.
    """

    def __init__(self, session) -> None:
        self.session = session
        self.base_url = f"https://v2.api.{self.session.env}.elucidata.io"
        self.base_url_auth = f"https://apis.{self.session.env}.elucidata.io/auth"
        self._cohort_details = None

    def add_to_cohort_single(
        self,
        cohort_path: str,
        repo_name: str,
        entity_id: list,
        entity_type: str,
    ):
        """
        Function to add dataset_id to cohort for single file structure repositories
        """
        print("Initializing process...")
        obj = CohortFileStandard()
        if not (entity_id and isinstance(entity_id, list)):
            raise InvalidParameterException(
                "entity_id(list) for single file structure repositories"
            )
        datatype_dict = {}
        with helpers.tqdm_joblib(tqdm(desc="Verifying Data", total=len(entity_id))):
            datatype_list = Parallel(n_jobs=20, require="sharedmem")(
                delayed(self._return_datatype_dict)(repo_name, i) for i in entity_id
            )
        for id in range(0, len(entity_id)):
            datatype_dict[entity_id[id]] = datatype_list[id]
        dataset_id = self._validate_dataset(entity_id, datatype_dict)
        with helpers.tqdm_joblib(
            tqdm(desc="Adding data to cohort", total=len(dataset_id))
        ):
            status_gct = Parallel(n_jobs=20, require="sharedmem")(
                delayed(self._gctfile)(repo_name, i, cohort_path) for i in dataset_id
            )
        with helpers.tqdm_joblib(
            tqdm(desc="Adding metadata to cohort", total=len(dataset_id))
        ):
            status_jpco = Parallel(n_jobs=20, require="sharedmem")(
                delayed(self._add_metadata)(repo_name, i, cohort_path)
                for i in dataset_id
            )
        dataset_id, deleted_id = obj._rollback_files(
            status_gct, status_jpco, repo_name, cohort_path, dataset_id
        )
        file_meta = helpers.make_path(cohort_path, "cohort.meta")
        tuple_list = self._get_datatype(repo_name, dataset_id, cohort_path)
        self._cohort_details = obj._update_metadata_single(
            repo_name, entity_type, dataset_id, deleted_id, file_meta, tuple_list
        )

    def _add_metadata(self, repo_key: str, dataset_id: str, local_path: str) -> None:
        """
        Function to add dataset level metadata to a cohort.
        Returns 0 on successful download, dataset_id on failure.
        """
        if not (repo_key and isinstance(repo_key, str)):
            raise InvalidParameterException("repo_id/repo_name")
        if not (dataset_id and isinstance(dataset_id, str)):
            raise InvalidParameterException("dataset_id")
        try:
            metadata = self._get_metadata(repo_key, dataset_id)
            file_name = f"{repo_key}_{dataset_id}.jpco"
            file_name = helpers.make_path(local_path, file_name)
            with open(file_name, "w") as outfile:
                json.dump(metadata, outfile)
            return 0
        except Exception:
            return dataset_id

    def _gctfile(self, repo_info: str, dataset_id: str, file_path: str) -> None:
        """
        Function to add gct file to a cohort
        Returns 0 on successful download, dtaset_id on failure.
        """
        try:
            if not (repo_info and isinstance(repo_info, str)):
                raise InvalidParameterException("repo_name/repo_id")
            if not (dataset_id and isinstance(dataset_id, str)):
                raise InvalidParameterException("dataset_id")
            ssl._create_default_https_context = ssl._create_unverified_context
            obj = OmixAtlas()
            download_dict = obj.download_data(repo_info, dataset_id, internal_call=True)
            url = (
                download_dict.get("data", {}).get("attributes", {}).get("download_url")
            )
            file_name = f"{repo_info}_{dataset_id}.gct"
            dest_path = helpers.make_path(file_path, file_name)
            urllib.request.urlretrieve(url, dest_path)
            return 0
        except Exception:
            return dataset_id

    def create_cohort(self, local_path: str, cohort_name: str, description: str) -> str:
        """
        This function is used to create a cohort
        Args:
            | local_path(str): local path to instantiate the cohort
            | cohort_name(str): identifier name for the cohort
            | description(str): description about the cohort
            | repo_key(str): Optional argument: repo_key(repo_name/repo_id) for the omixatlas to be added
            | entity_id(list): Optional argument: list of sample_id or dataset_id to be added to the cohort
        Returns:
            | A confirmation message on creation of cohort
        """
        if not (local_path and isinstance(local_path, str)):
            raise InvalidParameterException("local_path")
        if not (cohort_name and isinstance(cohort_name, str)):
            raise InvalidParameterException("cohort_name")
        if not (description and isinstance(description, str)):
            raise InvalidParameterException("description")
        if not os.path.exists(local_path):
            raise InvalidPathException
        if dot in cohort_name:
            raise InvalidCohortNameException(cohort_name)
        file_path = os.path.join(local_path, f"{cohort_name}.pco")
        user_id = self._get_user_id()
        os.makedirs(file_path)
        metadata = {
            "number_of_datasets": 0,
            "entity_id": {},
            "source_omixatlas": {},
            "description": description,
            "user_id": user_id,
            "date_created": str(datetime.datetime.now()),
            "version": COHORT_VERSION,
        }
        file_name = os.path.join(file_path, "cohort.meta")
        input = json.dumps(metadata)
        with open(file_name, "wb") as outfile:
            encoded_data = base64.b64encode(input.encode("utf-8"))
            outfile.write(encoded_data)
        logging.basicConfig(level=logging.INFO)
        logging.info("Cohort Created !")
        self._cohort_details = metadata
        return str(file_path)

    def _get_user_id(self):
        """
        Function to get user id
        """
        me_url = f"{self.base_url_auth}/users/me"
        details = self.session.get(me_url)
        error_handler(details)
        # user_id = details.json().get("data", {}).get("attributes", {}).get("user_id")
        user_id = int(details.json().get("data", {}).get("id"))
        return user_id

    def _get_datatype(self, repo_name: str, dataset_id: list, cohort_path: str) -> list:
        datatype_list = []
        for i in dataset_id:
            file_name = f"{repo_name}_{i}.jpco"
            file_name = helpers.make_path(cohort_path, file_name)
            if os.path.exists(file_name):
                with open(file_name, "r") as infile:
                    temp_df = json.load(infile)
                    data_type = temp_df.get("_source", {}).get("data_type")
                    dataset_tuple = [repo_name, data_type]
                    datatype_list.append(dataset_tuple)
        return datatype_list

    def _return_datatype_dict(self, repo_name: str, dataset_id: str):
        metadata = self._get_metadata(repo_name, dataset_id)
        datatype = metadata.get("_source", {}).get("data_type")
        return datatype

    def _get_metadata(self, repo_key: str, dataset_id: str) -> dict:
        """
        Function to return metadata for a dataset
        """
        obj = OmixAtlas()
        response_omixatlas = obj.omixatlas_summary(repo_key)
        data = response_omixatlas.get("data")
        index_name = data.get("indexes", {}).get("files")
        if index_name is None:
            raise paramException(
                title="Param Error", detail="Repo entered is not an omixatlas."
            )
        elastic_url = f"{obj.elastic_url}/{index_name}/_search"
        query = helpers.elastic_query(index_name, dataset_id)
        metadata = helpers.get_metadata(obj, elastic_url, query)
        return metadata

    def _validate_dataset(self, dataset_id: list, datatype_dict: dict) -> list:
        """
        Function to validate repo and datasets given as argument for adding to cohort
        """
        dataset_list = list(self._cohort_details["entity_id"].keys())
        valid_dataset_id = []
        dataset_id = list(set(dataset_id))
        for dataset in dataset_id:
            if dataset in dataset_list:
                logging.basicConfig(level=logging.INFO)
                logging.info(
                    f"The entity_id - {dataset} is already existing in the cohort."
                )
                continue
            datatype = datatype_dict[dataset]
            result = helpers.validate_datatype(datatype)
            if result:
                logging.basicConfig(level=logging.INFO)
                logging.info(
                    f"The entity_id - {dataset} has datatype - 'Single cell' that is not incorporated.\
                     Please contact Polly Support."
                )
                continue
            valid_dataset_id.append(dataset)
        if len(valid_dataset_id) == 0:
            raise InvalidDatasetException
        return valid_dataset_id

    def add_to_cohort_multiple(
        self,
        cohort_path: str,
        repo_name: str,
        dataset_id: str,
        sample_id: list,
        entity_type: str,
    ):
        """
        Function to add samples to cohort for multiple file structure repositories
        """
        if not (dataset_id and isinstance(dataset_id, str)):
            raise InvalidParameterException(
                "dataset_id(str) required for multiple file structure repository."
            )
        print("Initializing process...")
        metadata = self._get_metadata(repo_name, dataset_id)
        datatype = metadata.get("_source", {}).get("data_type")
        result = helpers.validate_datatype(datatype)
        if result:
            raise InvalidDatatypeException()
        dataset_list = list(self._cohort_details["entity_id"].keys())
        gct_file_name = f"{repo_name}_{dataset_id}.gct"
        jpco_file_name = f"{repo_name}_{dataset_id}.jpco"
        gct_path = helpers.make_path(cohort_path, gct_file_name)
        jpco_path = helpers.make_path(cohort_path, jpco_file_name)
        status = self._download_files_multiple(
            repo_name, dataset_id, dataset_list, cohort_path, gct_path, jpco_path
        )
        if status is None:
            raise TechnicalFaultException()
        download_flag = status
        data = parse(gct_path)
        cids = list(data.col_metadata_df.index)
        if sample_id is None or len(sample_id) == 0:
            sample_id = cids
        status_sample_id = []
        for i in sample_id:
            if i not in cids:
                status_sample_id.append(0)
            else:
                status_sample_id.append(1)
        file_meta = helpers.make_path(cohort_path, "cohort.meta")
        if 1 not in status_sample_id:
            print("Invalid samples, please try again.")
            if download_flag:
                os.remove(gct_path)
                os.remove(jpco_path)
        else:
            obj = CohortFileStandard()
            dataset_tuple = [
                repo_name,
                datatype,
            ]
            self._cohort_details = obj._update_metadata_multiple(
                repo_name,
                entity_type,
                dataset_id,
                sample_id,
                download_flag,
                file_meta,
                status_sample_id,
                dataset_tuple,
            )

    def _download_files_multiple(
        self, repo_name, dataset_id, dataset_list, local_path, gct_path, jpco_path
    ):
        download_flag = 0
        if dataset_id not in dataset_list:
            print("Adding data to cohort...")
            status_gct = self._gctfile(repo_name, dataset_id, local_path)
            print("Adding metadata to cohort...")
            status_jpco = self._add_metadata(repo_name, dataset_id, local_path)
            if not (status_gct == 0 and status_jpco == 0):
                if os.path.exists(gct_path):
                    os.remove(gct_path)
                if os.path.exists(jpco_path):
                    os.remove(jpco_path)
                return None
            download_flag = 1
        return download_flag

    def _return_omixatlas_name(self, repo_key: str):
        obj = OmixAtlas()
        response_omixatlas = obj.omixatlas_summary(repo_key)
        data = response_omixatlas.get("data")
        repo_name = data.get("repo_name")
        return repo_name

    def _get_entity_type(self, repo_name, dataset_id, cohort_constants):
        """
        Function to return the entity type for the hybrid omixatlases.
        Input: repo_name(str), dataset_id(str or list), cohort_constants(json)
        Returns: Entity_type('dataset' for single mapped repositories,\
            'sample' for multiple mapped repositories.)
        """
        if isinstance(dataset_id, list):
            with helpers.tqdm_joblib(
                tqdm(desc="Validating datasets", total=len(dataset_id))
            ):
                source_list = Parallel(n_jobs=20, require="sharedmem")(
                    delayed(self._return_data_source)(repo_name, dataset)
                    for dataset in dataset_id
                )
            # removing duplicate results
            unique_source = list(set(source_list))
            if len(unique_source) == 1 and unique_source[0] is None:
                # raising exception in case of wrong input
                raise TechnicalFaultException()
            # if more than one data sources, then check for compatibility
            if len(unique_source) > 1:
                entity_list = []
                for source in unique_source:
                    if source is None:
                        continue
                    entity_type = helpers.return_entity_type(
                        source.lower(), cohort_constants
                    )
                    if entity_type is not None:
                        entity_list.append(entity_type)
                entities = list(set(entity_list))
                if len(entities) > 1:
                    raise IncompatibleDataSource(unique_source)
                return entities[0]
            else:
                entity_type = helpers.return_entity_type(
                    unique_source[0].lower(), cohort_constants
                )
                return entity_type
        elif isinstance(dataset_id, str):
            data_source = self._return_data_source(repo_name, dataset_id)
            if data_source is None:
                raise TechnicalFaultException()
            entity_type = helpers.return_entity_type(
                data_source.lower(), cohort_constants
            )
            return entity_type
        else:
            # Invalid value for argument dataset_id
            raise InvalidParameterException("entity_id")

    def _return_data_source(self, repo_name, dataset_id):
        """
        Function to return the data source for a given repository name and dataset_id
        """
        try:
            metadata = self._get_metadata(repo_name, dataset_id)
            dataset_source = metadata.get("_source", {}).get("dataset_source")
            return dataset_source
        except Exception:
            print(
                f"Incorrect repo_key{{{repo_name}}} or dataset_id{{{dataset_id}}}. Please check the details"
            )



# Copyright 2017 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

################################## IMPORTANT #################################
# All user-facing functions in this package must be symlinked to the         #
# datacommons_pandas pkg. This is so that users do not need to import both   #
# libraries for pd support. Please keep the below imports in sync with the   #
# __init__.py in the datacommons_pandas/ dir, and add a symlink when         #
# creating a new file.                                                       #
# TODO: https://github.com/datacommonsorg/api-python/issues/149              #
##############################################################################

# Data Commons SPARQL query support
from datacommons.query import query

# Data Commons Python API
from datacommons.core import get_property_labels, get_property_values, get_triples
from datacommons.places import get_places_in, get_related_places, get_stats
from datacommons.populations import get_populations, get_observations, get_pop_obs, get_place_obs
from datacommons.stat_vars import get_stat_value, get_stat_series, get_stat_all

# Other utilities
from datacommons.utils import set_api_key



import datetime
import gzip
import io
import os
import shutil
import subprocess
import time
import typing as t
from contextlib import contextmanager
from pathlib import Path
from tempfile import NamedTemporaryFile, TemporaryDirectory, mkstemp

from .dp_types import NPath, log
from .utils import ON_WINDOWS


@contextmanager
def log_command(command: str) -> t.Generator[None, None, None]:
    """Log an internal process"""
    log.info(f"Starting {command}")
    yield
    log.info(f"Finished {command}")


@contextmanager
def create_temp_file(
    suffix: str, prefix: str = "datapane-temp-", mode: str = "w+b"
) -> t.Generator[NamedTemporaryFile, None, None]:
    """Creates a NamedTemporaryFile that doesn't disappear on .close()"""
    temp_file = NamedTemporaryFile(suffix=suffix, prefix=prefix, mode=mode, delete=False)
    try:
        yield temp_file
    finally:
        os.unlink(temp_file.name)


@contextmanager
def temp_fname(suffix: str, prefix: str = "datapane-temp-", keep: bool = False) -> t.Generator[str, None, None]:
    """Wrapper to generate a temporary filename only that is deleted on leaving context"""
    # TODO - return Path
    (in_f, in_f_name) = mkstemp(suffix=suffix, prefix=prefix)
    try:
        os.close(in_f)
        yield in_f_name
    finally:
        if os.path.exists(in_f_name) and not keep:
            os.unlink(in_f_name)


@contextmanager
def unix_compress_file(f_name: NPath, level: int = 6) -> t.Generator[str, None, None]:
    """(UNIX only) Return path to a compressed version of the input filename"""
    subprocess.run(["gzip", "-kf", f"-{level}", f_name], check=True)
    f_name_gz = f"{f_name}.gz"
    try:
        yield f_name_gz
    finally:
        os.unlink(f_name_gz)


@contextmanager
def unix_decompress_file(f_name: NPath) -> t.Generator[str, None, None]:
    """(UNIX only) Return path to a compressed version of the input filename"""
    subprocess.run(["gunzip", "-kf", f_name], check=True)
    f_name_gz = f"{f_name}.gz"
    try:
        yield f_name_gz
    finally:
        os.unlink(f_name_gz)


@contextmanager
def compress_file(f_name: NPath, level: int = 6) -> t.Generator[str, None, None]:
    """(X-Plat) Return path to a compressed version of the input filename"""
    f_name_gz = f"{f_name}.gz"
    with open(f_name, "rb") as f_in, gzip.open(f_name_gz, "wb", compresslevel=level) as f_out:
        shutil.copyfileobj(f_in, f_out)
    try:
        yield f_name_gz
    finally:
        # NOTE - disable on windows temporarily
        if not ON_WINDOWS:
            os.unlink(f_name_gz)


def inmemory_compress(content: t.BinaryIO) -> t.BinaryIO:
    """(x-plat) Memory-based gzip compression"""
    content.seek(0)
    zbuf = io.BytesIO()
    with gzip.GzipFile(mode="wb", fileobj=zbuf, mtime=0.0) as zfile:
        zfile.write(content.read())
    zbuf.seek(0)
    return zbuf


@contextmanager
def temp_workdir() -> t.Generator[str, None, None]:
    """Set working dir to a tempdir for duration of context"""
    with TemporaryDirectory() as tmp_dir:
        curdir = os.getcwd()
        os.chdir(tmp_dir)
        try:
            yield None
        finally:
            os.chdir(curdir)


@contextmanager
def pushd(directory: NPath, pre_create: bool = False, post_remove: bool = False) -> t.Generator[None, None, None]:
    """Switch dir and push it onto the (call-)stack"""
    directory = Path(directory)
    cwd = os.getcwd()
    log.debug(f"[cd] {cwd} -> {directory}")
    if not directory.exists() and pre_create:
        Path(directory).mkdir(parents=True)
    os.chdir(directory)
    try:
        yield
    finally:
        log.debug(f"[cd] {cwd} <- {directory}")
        os.chdir(cwd)
        if post_remove:
            shutil.rmtree(directory, ignore_errors=True)


def get_filesize(filename: Path) -> int:
    return filename.stat().st_size


def walk_path(path: Path) -> t.Iterable[Path]:
    for p in path.rglob("*"):
        if not p.is_dir():
            yield p


def unixtime() -> int:
    return int(time.time())


def timestamp(x: t.Optional[datetime.datetime] = None) -> str:
    """Return ISO timestamp for a datetime"""
    x = x or datetime.datetime.utcnow()
    return f'{x.isoformat(timespec="seconds")}{"" if x.tzinfo else "Z"}'



def generate_swagger_html(spec: str, js_url: str, css_url: str) -> str:
    """
    Generate Swagger UI HTML page

    Parameters
    ----------
    spec: str
        The OpenAPI spec in the JSON format
    js_url: str
        The URL to the Swagger UI JavaScript file
    css_url: str
        The URL to the Swagger UI CSS file
    """
    return f"""
<!DOCTYPE html>
<html>
<head>
    <meta charset="UTF-8">
    <title>Swagger UI</title>
    <meta
      http-equiv="Cache-control"
      content="no-cache, no-store, must-revalidate"
    />
    <link rel="stylesheet" type="text/css" href="{css_url}">
</head>

<body>
    <div id="swagger-ui">
        Loading...
    </div>
</body>

<script src="{js_url}"></script>

<script>
  var swaggerUIOptions = {{
    dom_id: "#swagger-ui",
    docExpansion: "list",
    deepLinking: true,
    filter: true,
    layout: "BaseLayout",
    showExtensions: true,
    showCommonExtensions: true,
    spec: JSON.parse(`
        {spec}
    `.trim()),
    presets: [
      SwaggerUIBundle.presets.apis,
      SwaggerUIBundle.SwaggerUIStandalonePreset
    ],
    plugins: [
      SwaggerUIBundle.plugins.DownloadUrl
    ]
  }}

  var ui = SwaggerUIBundle(swaggerUIOptions)
</script>
</html>
            """.strip()




from pytest import raises

from glom import glom, Path, S, T, A, PathAccessError, GlomError, BadSpec, Or, Assign, Delete
from glom import core

def test_list_path_access():
    assert glom(list(range(10)), Path(1)) == 1


def test_path():
    _obj = object()
    target = {'a': {'b.b': [None, {_obj: [None, None, 'd']}]}}

    assert glom(target, Path('a', 'b.b', 1, _obj, -1)) == 'd'


def test_empty_path_access():
    target = {}

    assert glom(target, Path()) is target
    assert glom(target, (Path(), Path(), Path())) is target

    dup_dict = glom(target, {'target': Path(),
                             'target2': Path()})
    dup_dict['target'] is target
    dup_dict['target2'] is target


def test_path_t_roundtrip():
    # check that T repr roundrips
    assert repr(T['a'].b.c()) == "T['a'].b.c()"
    assert repr(T[1:]) == "T[1:]"
    assert repr(T[::3, 1:, 1:2, :2:3]) == "T[::3, 1:, 1:2, :2:3]"

    # check that Path repr roundtrips
    assert repr(Path('a', 1, 'b.b', -1.0)) == "Path('a', 1, 'b.b', -1.0)"

    # check that Path repr roundtrips when it contains Ts
    assert repr(Path(T['a'].b, 'c', T['d'].e)) == "Path(T['a'].b, 'c', T['d'].e)"

    # check that T instances containing Path access revert to repring with Path
    assert repr(Path(T['a'].b, 'c', T['d'].e).path_t) == "Path(T['a'].b, 'c', T['d'].e)"

    # check that Paths containing only T objects reduce to a T (joining the T objects)
    assert repr(Path(T['a'].b, T.c())) == "T['a'].b.c()"

    # check that multiple nested paths reduce
    assert repr(Path(Path(Path('a')))) == "Path('a')"

    # check builtin repr
    assert repr(T[len]) == 'T[len]'
    assert repr(T.func(len, sum)) == 'T.func(len, sum)'

    # check * and **
    assert repr(T.__star__().__starstar__()) == 'T.__star__().__starstar__()'
    assert repr(Path('a', T.__star__().__starstar__())) == "Path('a', T.__star__().__starstar__())"


def test_path_access_error_message():

    # test fuzzy access
    with raises(GlomError) as exc_info:
        glom({}, 'a.b')
    assert ("PathAccessError: could not access 'a', part 0 of Path('a', 'b'), got error: KeyError"
            in exc_info.exconly())
    ke = repr(KeyError('a'))  # py3.7+ changed the keyerror repr
    assert repr(exc_info.value) == "PathAccessError(" + ke + ", Path('a', 'b'), 0)"

    # test multi-part Path with T, catchable as a KeyError
    with raises(KeyError) as exc_info:
        # don't actually use glom to copy your data structures please
        glom({'a': {'b': 'c'}}, Path('a', T.copy(), 'd'))
    assert ("PathAccessError: could not access 'd', part 3 of Path('a', T.copy(), 'd'), got error: KeyError"
            in exc_info.exconly())
    ke = repr(KeyError('d'))  # py3.7+ changed the keyerror repr
    assert repr(exc_info.value) == "PathAccessError(" + ke + ", Path('a', T.copy(), 'd'), 3)"

    # test AttributeError
    with raises(GlomError) as exc_info:
        glom({'a': {'b': 'c'}}, Path('a', T.b))
    assert ("PathAccessError: could not access 'b', part 1 of Path('a', T.b), got error: AttributeError"
            in exc_info.exconly())
    ae = repr(AttributeError("'dict' object has no attribute 'b'"))
    assert repr(exc_info.value) == "PathAccessError(" + ae + ", Path(\'a\', T.b), 1)"


def test_t_picklability():
    import pickle

    class TargetType(object):
        def __init__(self):
            self.attribute = lambda: None
            self.attribute.method = lambda: {'key': lambda x: x * 2}

    spec = T.attribute.method()['key'](x=5)

    rt_spec = pickle.loads(pickle.dumps(spec))
    assert repr(spec) == repr(rt_spec)

    assert glom(TargetType(), spec) == 10

    s_spec = S.attribute
    assert repr(s_spec) == repr(pickle.loads(pickle.dumps(s_spec)))


def test_t_subspec():
    # tests that arg-mode is a min-mode, allowing for
    # other specs to be embedded inside T calls
    data = [
        {'id': 1},
        {'pk': 1}]

    get_ids = (
        S(id_type=int),
        [S.id_type(Or('id', 'pk'))])

    assert glom(data, get_ids) == [1, 1]

    data = {'a': 1, 'b': 2, 'c': 3}

    # test that "shallow" data structures translate as-is
    get_vals = (
        S(seq_type=tuple),
        S.seq_type([T['a'], T['b'], Or('c', 'd')])
    )

    assert glom(data, get_vals) == (1, 2, 3)


def test_a_forbidden():
    with raises(BadSpec):
        A()  # cannot assign to function call
    with raises(BadSpec):
        glom(1, A)  # cannot assign without destination


def test_s_magic():
    assert glom(None, S.test, scope={'test': 'value'}) == 'value'

    with raises(PathAccessError):
        glom(1, S.a)  # ref to 'a' which doesn't exist in scope

    with raises(PathAccessError):
        glom(1, A.b.c)

    return


def test_path_len():

    assert len(Path()) == 0
    assert len(Path('a', 'b', 'c')) == 3
    assert len(Path.from_text('1.2.3.4')) == 4

    assert len(Path(T)) == 0
    assert len(Path(T.a.b.c)) == 3
    assert len(Path(T.a()['b'].c.d)) == 5


def test_path_getitem():
    path = Path(T.a.b.c)

    assert path[0] == Path(T.a)
    assert path[1] == Path(T.b)
    assert path[2] == Path(T.c)
    assert path[-1] == Path(T.c)
    assert path[-2] == Path(T.b)

    with raises(IndexError, match='Path index out of range'):
        path[4]

    with raises(IndexError, match='Path index out of range'):
        path[-14]
    return


def test_path_slices():
    path = Path(T.a.b, 1, 2, T(test='yes'))

    assert path[::] == path

    # positive indices
    assert path[3:] == Path(2, T(test='yes'))
    assert path[1:3] == Path(T.b, 1)
    assert path[:3] == Path(T.a.b, 1)

    # positive indices backwards
    assert path[2:1] == Path()

    # negative indices forward
    assert path[-1:] == Path(T(test='yes'))
    assert path[:-2] == Path(T.a.b, 1)
    assert path[-3:-1] == Path(1, 2)

    # negative indices backwards
    assert path[-1:-3] == Path()

    # slicing and stepping
    assert path[1::2] == Path(T.b, 2)


def test_path_values():
    path = Path(T.a.b, 1, 2, T(test='yes'))

    assert path.values() == ('a', 'b', 1, 2, ((), {'test': 'yes'}))

    assert Path().values() == ()


def test_path_items():
    path = Path(T.a, 1, 2, T(test='yes'))

    assert path.items() == (('.', 'a'),
                            ('P', 1), ('P', 2),
                            ('(', ((), {'test': 'yes'})))

    assert Path().items() == ()


def test_path_star():
    val = {'a': [1, 2, 3]}
    assert glom(val, 'a.*') == [1, 2, 3]
    val['a'] = [{'b': v} for v in val['a']]
    assert glom(val, 'a.*.b') == [1, 2, 3]
    assert glom(val, T['a'].__star__()['b']) == [1, 2, 3]
    assert glom(val, Path('a', T.__star__(), 'b')) == [1, 2, 3]
    # multi-paths eat errors
    assert glom(val, Path('a', T.__star__(), T.b)) == []
    val = [[[1]]]
    assert glom(val, '**') == [val, [[1]], [1], 1]
    val = {'a': [{'b': [{'c': 1}, {'c': 2}, {'d': {'c': 3}}]}], 'c': 4}
    assert glom(val, '**.c') == [4, 1, 2, 3]
    assert glom(val, 'a.**.c') == [1, 2, 3]
    assert glom(val, T['a'].__starstar__()['c']) == [1, 2, 3]
    assert glom(val, 'a.*.b.*.c') == [[1, 2]]
    # errors
    class ErrDict(dict):
        def __getitem__(key): 1/0
    assert ErrDict(val).keys()  # it will try to iterate
    assert glom(ErrDict(val), '**') == [val]
    assert glom(ErrDict(val), '*') == []
    # object access
    class A:
        def __init__(self):
            self.a = 1
            self.b = {'c': 2}
    val = A()

    assert glom(val, '*') == [1, {'c': 2}]
    assert glom(val, '**') == [val, 1, {'c': 2}, 2]


def test_star_broadcast():
    val = {'a': [1, 2, 3]}
    assert glom(val, Path.from_text('a.*').path_t + 1) == [2, 3, 4]
    val = {'a': [{'b': [{'c': 1}, {'c': 2}, {'c': 3}]}]}
    assert glom(val, Path.from_text('**.c').path_t + 1) == [2, 3, 4]


def test_star_warning():
    '''check that the default behavior is as expected; this will change when * is default on'''
    assert core.PATH_STAR is True
    try:
        core.PATH_STAR = False
        assert glom({'*': 1}, '*') == 1
        assert Path._STAR_WARNED
    finally:
        core.PATH_STAR = True

def test_path_eq():
    assert Path('a', 'b') == Path('a', 'b')
    assert Path('a') != Path('b')

    assert Path() != object()


def test_path_eq_t():
    assert Path(T.a.b) == T.a.b
    assert Path(T.a.b.c) != T.a.b


def test_startswith():
    ref = T.a.b[1]

    assert Path(ref).startswith(T)
    assert Path(ref).startswith(T.a.b)
    assert Path(ref).startswith(ref)
    assert Path(ref).startswith(ref.c) is False

    assert Path('a.b.c').startswith(Path())
    assert Path('a.b.c').startswith('a.b.c')

    with raises(TypeError):
        assert Path('a.b.c').startswith(None)

    return


def test_from_t_identity():
    ref = Path(T.a.b)
    assert ref.from_t() == ref
    assert ref.from_t() is ref


def test_t_dict_key():
    target = {'a': 'A'}
    assert glom(target, {T['a']: 'a'}) == {'A': 'A'}


def test_t_arithmetic():
    t = 2
    assert glom(t, T + T) == 4
    assert glom(t, T * T) == 4
    assert glom(t, T ** T) == 4
    assert glom(t, T / 1) == 2
    assert glom(t, T % 1) == 0
    assert glom(t, T - 1) == 1
    assert glom(t, T & T) == 2
    assert glom(t, T | 1) == 3
    assert glom(t, T ^ T) == 0
    assert glom(2, ~T) == -3
    assert glom(t, -T) == -2


def test_t_arithmetic_reprs():
    assert repr(T + T) == "T + T"
    assert repr(T + (T / 2 * (T - 5) % 4)) == "T + (T / 2 * (T - 5) % 4)"
    assert repr(T & 7 | (T ^ 6)) == "T & 7 | (T ^ 6)"
    assert repr(-(~T)) == "-(~T)"


def test_t_arithmetic_errors():
    with raises(PathAccessError, match='zero'):
        glom(0, T / 0)

    with raises(PathAccessError, match='unsupported operand type'):
        glom(None, T / 2)

    return


def test_t_dunders():
    with raises(AttributeError) as exc_info:
        T.__name__

    assert 'use T.__("name__")' in str(exc_info.value)

    assert glom(1, T.__('class__')) is int


def test_path_cache():
    assert Path.from_text('a.b.c') is Path.from_text('a.b.c')
    pre = Path._MAX_CACHE
    Path._MAX_CACHE = 0
    assert Path.from_text('d.e.f') is not Path.from_text('d.e.f')



from __future__ import annotations

import json
import os
import pickle
import typing as t
from pathlib import Path

import pandas as pd

from datapane.client import DPClientError, log
from datapane.common import PKL_MIMETYPE, ArrowFormat, NPath
from datapane.common.datafiles import DFFormatterCls, df_ext_map
from datapane.common.df_processor import to_df

from .common import DPTmpFile, do_download_file
from .dp_object import DPObjectRef


def save_df(df: pd.DataFrame) -> DPTmpFile:
    """Export a df for uploading"""
    fn = DPTmpFile(ArrowFormat.ext)

    # create a copy of the df to process
    df = to_df(df)
    if df.size == 0:
        raise DPClientError("Empty DataFrame provided")

    # process_df called in Arrow.save_file
    # process_df(df)
    ArrowFormat.save_file(fn.name, df)
    log.debug(f"Saved df to {fn} ({os.path.getsize(fn.file)} bytes)")
    return fn


class File(DPObjectRef):
    """
    Files are files that can be uploaded and downloaded for use in your apps,
    for instance trained models, datasets, and (pickled) Python objects.
    They are generally large, but can be any size.

    Attributes:
        content_type: the file content-type
        size_bytes: the file size
        num_rows: number of rows in the file (if a dataframe)
        num_colums: number of colums in the file (if a dataframe)
        cells: number of cells in the file (if a dataframe)
        overwrite: overwrite the file in project

    ..tip:: Use the static methods to create `Files` rather than the constructor
    """

    endpoint: str = "/files/"

    @classmethod
    def upload_df(cls, df: pd.DataFrame, **kwargs) -> File:
        """
        Create a file containing the dataframe provided

        Args:
            df: The pandas dataframe to upload as a File

        Returns:
            An instance of the created `File` object
        """
        with save_df(df) as fn:
            return cls.post_with_files(file=fn.file, **kwargs)

    @classmethod
    def upload_file(cls, fn: NPath, **kwargs) -> File:
        """
        Create a file containing the contents of the file provided

        Args:
            fn: Path to the file to upload as a File

        Returns:
            An instance of the created `File` object
        """
        return cls.post_with_files(file=Path(fn), **kwargs)

    @classmethod
    def upload_obj(cls, data: t.Any, **kwargs) -> File:
        """
        Create a file containing the contents of the Python object provided,
        the object may be pickled or converted to JSON before storing.

        Args:
            data: Python object to upload as a File

        Returns:
            An instance of the created `File` object
        """
        # import here as a very slow module due to nested imports
        from .obsolete.files import save

        fn = save(data)
        return cls.post_with_files(file=fn.file, **kwargs)

    def download_df(self) -> pd.DataFrame:
        """
        Download the file and return it as a Dataframe

        Returns:
            A pandas dataframe generated from the file
        """
        with DPTmpFile(ArrowFormat.ext) as fn:
            do_download_file(self.data_url, fn.name)
            return ArrowFormat.load_file(fn.name)

    def download_file(self, fn: NPath) -> None:
        """
        Download the file to the file provided

        Args:
            fn: Path representing the location to save the file
        """
        fn = t.cast(Path, fn)

        def get_export_format() -> DFFormatterCls:
            filename = t.cast(Path, fn)
            ext = filename.suffix
            if ext not in df_ext_map:
                raise DPClientError(
                    f"Extension {ext} not valid for exporting table. Must be one of {', '.join(df_ext_map.keys())}"
                )
            return df_ext_map[ext]

        # If file is of arrow type, export it. Otherwise, download the url directly.
        if self.content_type == ArrowFormat.content_type:
            # download as a arrow->df, then export on the client
            df = self.download_df()
            fmt = get_export_format()
            fmt.save_file(str(fn), df)
        else:
            do_download_file(self.data_url, fn)

    def download_obj(self) -> t.Any:
        """
        Download the file and return it as a Python object

        Returns:
            The object created by deserialising the File (either via Pickle or JSON decoding)
        """
        with DPTmpFile(".obj") as fn:
            do_download_file(self.data_url, fn.name)
            # In the case that the original object was a Python object or bytes-like object,
            # the downloaded obj will be a pickle which needs to be unpickled.
            # Otherwise it's a stringified JSON object (e.g. an Altair plot) that can be returned as JSON.
            if self.content_type == PKL_MIMETYPE:
                with fn.file.open("rb") as fp:
                    return pickle.load(fp)
            else:
                return json.loads(fn.file.read_text())



class _FrozenDict(dict):
    """
    A dictionary that can be used as a key in another dictionary.

    This is needed because the default dict implementation is not hashable.
    The only usage for this right now is to store dicts as part of the Router key.
    The implementation only takes into consideration the keys of the dictionary.

    MAINTENANCE: this is a temporary solution until we refactor the route key into a class.
    """

    def __hash__(self):
        return hash(frozenset(self.keys()))
